# File: evmBBB_VIRCAMtests.tcl
# Purpose: VIRCAM specific EVM tests
# Author: Stephane Marteau
#

# $Id: evm_BBB_VIRCAMtests.tcl 350144 2021-12-09 09:40:19Z smysore $

# OB data - global array containing values of all needed parameters for all templates
# of all OBs.
#
# OBdata(<ob_id>,<tpl_idx>,<keyword>) = <value>
#

global OBdata
array set OBdata {}

#####################  VIRCAM TEMPLATES  ####################
global evmVIRCAMTpl
array set evmVIRCAMTpl {}

# VIRCAM acquisition templates
set evmVIRCAMTpl(TILE,acq)     VIRCAM_img_acq_tile
set evmVIRCAMTpl(QUICK,acq)    VIRCAM_img_acq_quick

# VIRCAM science templates
set evmVIRCAMTpl(TILE,sci,1)   VIRCAM_img_obs_tile1
set evmVIRCAMTpl(TILE,sci,2)   VIRCAM_img_obs_tile3
set evmVIRCAMTpl(TILE,sci,3)   VIRCAM_img_obs_tile6
set evmVIRCAMTpl(TILE,sci,4)   VIRCAM_img_obs_tile6sky

set evmVIRCAMTpl(QUICK,sci,1)  VIRCAM_img_obs_paw
set evmVIRCAMTpl(QUICK,cal,1)  VIRCAM_img_cal_std
set evmVIRCAMTpl(QUICK,cal,2)  VIRCAM_img_cal_illumination

#### DIT Limits
global evmDITLimits
array set evmDITLimits {}

set evmDITLimits(saturate,Z)      600
set evmDITLimits(saturate,Y)      170
set evmDITLimits(saturate,J)       50
set evmDITLimits(saturate,NB118) 2000
set evmDITLimits(saturate,NB980) 2000
set evmDITLimits(saturate,H)       10
set evmDITLimits(saturate,Ks)      12

set evmDITLimits(RONlimited,Z)     20
set evmDITLimits(RONlimited,Y)      5
set evmDITLimits(RONlimited,J)      1.5
set evmDITLimits(RONlimited,NB118) 60
set evmDITLimits(RONlimited,NB980) 60

#############################################################
##
## Main EVM Procedures
##
#############################################################

# Procedure:getFilterChangeTime
#
# Returns the time needed to change from one filter to another one.
# This time depends on the relative position of the 2 filters in the
# VIRCAM filter wheel.
#
# Arguments:
#  - filter1: name of current filter
#  - filter2: name of filter to switch to...
#
# Returns:
#  time to switch from filter1 to filter2 (in seconds)
#
proc getFilterChangeTime { filter1 filter2 } {

    # Overhead table (nb of positions to skip --> time in seconds)
    set time(0) 0
    set time(1) 20.5
    set time(2) 27
    set time(3) 33
    set time(4) 40

    # ORDERED List of VIRCAM filters
    set filterList {SUNBLIND NB980 H Ks J Y Z NB118}

    if { [set idx1 [lsearch -exact $filterList $filter1]] == -1 || \
	[set idx2 [lsearch -exact $filterList $filter2]] == -1 } {

	return 0

    }

    set nbFilters [llength $filterList]

    # Since this is a wheel we compute the separation from both ends
    # of the list and take the "shortest path"
    set separation1 [expr abs($idx2 - $idx1)]
    set separation2 [expr $nbFilters - $separation1]
    set separation [min $separation1 $separation2]

    # Get the overhead from the above table
    return $time($separation)

}

#### VIRCAM Level 1 Checks - General tests

proc checkVIRCAMLevel1 { object } {

    # Global variables
    global scriptingBox
    global OBdata Phase1data

    # Get 'object type' (first element of object list)
    set objType [lindex $object 0]

    set listOfObsBlocks {}
    set isContainer false

    # If the object type is a list, then we are dealing
    # with a container: the actual type has to be fetched
    # deeper inside the object list
    if { [llength $objType] > 1 } {

	set containerInfo [lindex $object 0]

	set objType [lindex $containerInfo 0] ; # Type of container
	set containerType [lindex $containerInfo 0] ; # Type of container
	set containerID [lindex $containerInfo 1] ; # Container ID
	set objRef  [lindex $containerInfo 2] ; # Object reference (Java handle)
	set objSelected [lindex $containerInfo 3] ; # Is container selected or not
	set containerName [evmName $objRef]

	set listOfObsBlocks [lindex $object 1]
	set isContainer true

    } else {

	set ob_id [lindex $object 1] ; # OB ID
	set objRef [lindex $object 2] ; # Object reference (Java handle)
	set objSelected [lindex $object 3] ; # Is object part of selection or not
	set obName $OBdata($ob_id,name)

	set listOfObsBlocks [list $object]

    }

    # Determine observing mode using first OB
    set ob [lindex $listOfObsBlocks 0]
    set ob_id [lindex $ob 1]
    set obRunMode $OBdata($ob_id,runMode)
    set runID $OBdata($ob_id,runID)

    # Run letter
    if { ![regexp {([A-Z])\)$} $runID dummy runLetter] } {
	regexp {[0-9][0-9][0-9]$} $runID runLetter
	set runLetter [format "%c" [expr [string trimleft $runLetter 0] + 64]]
    }

    #
    # Checks at container level
    #

    # For VM: check that containers are not used [General Check]
    if { $obRunMode == "v" && $isContainer && $objSelected } {
	set msg "It is not possible to use OB containers in Visitor Mode!"
	issueError $containerID $msg $runID
    }

    # Sanity check for time delays in TIME-LINK containers [General Check]
    if { $obRunMode == "s" && $objType == "TIME_LINK"  && $objSelected } {
	checkTimeLinkSanity $objRef $containerID 1
    }

    #
    # Loop on all OBs

    # list of OB absolute time interval lists
    set concat_OB_absTI_list {}

    #vars for weather constraints:
    set concat_seeing_list {}
    set concat_fli_list {}
    set concat_transp_list {}
    set concat_twilight_list {}

    #list of ra,dec pairs
    set ra_dec_list {}

    set totalTimeConcat 0

    set obIdx 0
    #seconds in a standard year (365.242199 * 86,400)
    set secsInYear 31556926

    # vars for LST checks
    set concat_LST_list {}
    set hasLSTinConcat 0

    foreach ob $listOfObsBlocks {

	incr obIdx

	set isOBSelected [lindex $ob 3]

	set ob_id [lindex $ob 1]
	set obRunMode $OBdata($ob_id,runMode)

	set objType [lindex $ob 0]
	set objRef [lindex $ob 2]

	# Get OB name, run ID and run mode
	set obName $OBdata($ob_id,name)

	#
	# DETERMINE: obType (Observation/Calibration)
	#
	set obType $OBdata($ob_id,type)

	#
	# If OB = Calib, we skip the rest of the tests (irrelevant)
	#
	if { $obType == "C" } {
	    set msg "Calibration OB: no check performed"
	    issueInfo $ob_id $msg
	    continue
	}

	set execTime [evmExecutionTime $objRef]

	# Compute the total execution time of selected OBs/containers
	# (use global variable for use outside of main EVM loop)
	set OBdata(totalExecutionTime) [expr $OBdata(totalExecutionTime) + $execTime]

	# collect execution times and weather conditions/constraints:
	set csSeeing  $OBdata($ob_id,csSeeing)
	set csFli $OBdata($ob_id,csFLI)
	set csTransp $OBdata($ob_id,csTransp)
	set csTwilight $OBdata($ob_id,csTwilight)

	# collect ra,dec pairs
	set ra [expr $OBdata($ob_id,ra_marcsec) / 1000.]
	set dec [expr $OBdata($ob_id,dec_marcsec) / 1000.]
	set ra_dec [list $ra $dec]

	# vv SETUP FOR LATER CONTAINER-LEVEL CHECKS vv

	if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && $objSelected } {
	    # add to total exectime for a check (outside foreach loop)
	    set totalTimeConcat [expr $totalTimeConcat + [evmExecutionTime $objRef]]

	    set tis [evmTimeIntervals $objRef]
	    #create a list of OB {startTime endTime} pairs
	    set obTimeIntervalsList {}
	    if { [llength $tis] } {
		foreach ti $tis {
		    set startTime [evmStartTime $ti]
		    set endTime [evmEndTime $ti]
		    lappend obTimeIntervalsList [list $startTime $endTime]
		}
	    }

	    #if any absolute time intervals are defined, then append the sorted list to the absolute time intervals list of the
	    #concatenation
	    if { [llength $obTimeIntervalsList] != 0} {
		#sort this list of time intervals since they may not be in chronological order (use custom sort obTIsort)
		set obTimeIntervalsList [lsort -command obTIsort $obTimeIntervalsList]
		# collect OB absolute time interval lists
		lappend concat_OB_absTI_list $obTimeIntervalsList
	    }

	    # collect WEATHER CONSTRAINTS
	    lappend concat_seeing_list $csSeeing
	    lappend concat_fli_list $csFli
	    lappend concat_transp_list $csTransp
	    lappend concat_twilight_list $csTwilight

	    #collect ra,dec pairs
	    lappend ra_dec_list $ra_dec
	}

	if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && $objSelected } {
	    if { $OBdata($ob_id,LST,nbIntervals) > 0 } {
		set hasLSTinConcat 1
	    }
	}

	set ob_LSTlist {}

	for {set ti 1} {$ti <= $OBdata($ob_id,LST,nbIntervals)} {incr ti} {

	    set lst_min [expr $OBdata($ob_id,LST,$ti,start)/3600.] ; # hours
	    set lst_max [expr $OBdata($ob_id,LST,$ti,end)/3600.] ; # hours

	    #collect this here for use later in the LST checks for concats
	    if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && $objSelected } {
		#collect all the time intervals for this OB , then see end of for loop
		lappend ob_LSTlist [list $lst_min $lst_max]
	    }

	} ; #end for

	# sort and add the intervals for each OB along with the OBidx; now see code at the end of OB loop (LST check for concat)
	lappend concat_LST_list [list $obIdx [lsort -command obTIsort $ob_LSTlist]]

	# ^^ END OF SETUP FOR LATER CONTAINER-LEVEL CHECKS ^^

	# vv Checks below are for OB-level only in this loop vv

	# Is the OB actually part of the selection? If not, skip the OB-level checks...
	if { !$isOBSelected } { continue }

	# Check OB name for unallowed characters [General Check]:
	if { [regexp {^[a-zA-Z0-9+_ \.-]+$} $obName] == 0 } {

	    set msg "$ob_id: OB with unallowed characters in its name. Please fix it.\n            (allowed characters: letters, digits, space, + - _ . )"
	    issueError $ob_id $msg $runID

	}

	# [General Check] Check OB name for "No name":
	if { [regexp -nocase {No[_\. -]*name} $obName] != 0 } {

	    set msg "Please provide an OB name which does not contain 'No name'."
	    issueError $ob_id $msg $runID

	}

	# [General Check] ToO/RRM prefix:
	if { $Phase1data(tooFlag) == 1 } {

	    if { ![regexp -nocase {^(TOO_|RRM_)} $OBdata($ob_id,name)] } {
		set msg "This is a ToO/RRM run, please add the prefix 'TOO_' or 'RRM_' to the name of the OB\n  depending on the kind of OB."
		if { $obRunMode == "s" } {
		    issueError $ob_id $msg $runID
		} else {
		    issueWarning $ob_id $msg
		}
	    }

	} else {

	    if { [regexp -nocase {^(TOO_)} $OBdata($ob_id,name)] } {
		set msg "This is not a ToO run, please remove the prefix 'TOO_' from OB name."
		if { $obRunMode == "s" } {
		    issueError $ob_id $msg $runID
		} else {
		    issueWarning $ob_id $msg
		}
	    }
	    if { [regexp -nocase {^(RRM_)} $OBdata($ob_id,name)] } {
		set msg "This is not a ToO/RRM run, please remove the prefix 'RRM_' from OB name."
		if { $obRunMode == "s" } {
		    issueError $ob_id $msg $runID
		} else {
		    issueWarning $ob_id $msg
		}
	    }

	}

	#
	#  [General Check] Check execution time of OB
	#
	# (execTime has been set at the beginning of the OB loop)

	if { $execTime > 3900 } {

	    set printedTime [formatPrintedTime $execTime]

	    set msg "Execution time is greater than 1h05min (found $printedTime)"
	    if { $obRunMode == "s" } {
		issueErrorWaiver $ob_id $msg $runID 3
	    } else {
		issueWarning $ob_id $msg
	    }
	}

	# Check differential tracking ERROR (SM)
	if { $obRunMode == "s" && ( $OBdata($ob_id,diff_ra) != 0 || $OBdata($ob_id,diff_dec) != 0 || \
		$OBdata($ob_id,0,TEL.TARG.ADDVELALPHA) != 0 || $OBdata($ob_id,0,TEL.TARG.ADDVELDELTA) != 0 ) } {

	    set msg "Differential tracking values are not set to zero!"
	    issueError $ob_id $msg $runID

	}

	# Check for over-specified seeing constraint ERROR (SM)
	if { $obRunMode == "s" && ( $OBdata($ob_id,type) != "C"  && $OBdata($ob_id,csSeeing) < 0.6 ) } {

	    set msg "Seeing constraint is smaller than 0.6\""
	    issueError $ob_id $msg $runID

	}

	# [General Check] Check target name, coordinates and airmass constraints including zenith avoidance
	set targetOK [checkTarget $ob_id $runID]

	#
	# [General Check] Time Constraints Check
	#

	#Get the time intervals set
	set tis [evmTimeIntervals $objRef]
	if { [llength $tis] } {

	    #Check if ToO/RRM (but not TOO_MOV_) OB has absolute time intervals, and exit if error
	    if { $Phase1data(tooFlag) == 1 && $obRunMode == "s" && ![regexp -nocase {^(TOO_|RRM_)?MOV_} $OBdata($ob_id,name)] } {
		set msg "Please remove absolute time windows because this is a TOO/RRM observation. \n"
		append msg "        The time constraint for execution is provided when the trigger is issued."

		issueError $ob_id $msg $runID

		continue
	    }

	    set tt 0
	    set currentTime [clock seconds]
	    set earliestStartTime $currentTime
	    set latestEndTime $currentTime
	    # in case the returned set of time intervals in not ordered, get the earliest
	    foreach ti $tis {
		set startTime [evmStartTime $ti]
		set endTime [evmEndTime $ti]
		set duration [expr $endTime-$startTime]

		# Zero time interval [General Check]
		if { $duration == 0 } {
		    set msg "OB cannot have a defined absolute time interval of zero hours."
		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    } else {
			issueWarning $ob_id $msg
		    }
		} elseif { $duration <= 1800. } {
		    # Short time interval [General Check]
		    set msg "An absolute time interval for this OB is less than 30 minutes. Please\n"
		    set msg "$msg consider extending it in order to increase the chances for this OB to be scheduled."
		    issueWarning $ob_id $msg
		} elseif { $duration > $secsInYear } {
		    # Long time interval [General Check]
		    set msg "The OB cannot have an absolute time interval longer than 1 year."
		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    } else {
			issueWarning $ob_id $msg
		    }
		}
		# For Expired time intervals [General Check]
		if { $tt == 0 } {
		    set earliestStartTime $startTime
		    set latestEndTime $endTime
		} else {
		    if { $startTime < $earliestStartTime } {
			set earliestStartTime $startTime
		    }
		    if { $endTime > $latestEndTime } {
			set latestEndTime $endTime
		    }
		}
		incr tt
	    }
	    # - If the minimum date/time is in the past (before now)
	    if { $latestEndTime <= $currentTime } {
		set msg "Absolute time intervals must refer to some time in the future."
		if { $obRunMode == "s" } {
		    issueError $ob_id $msg $runID
		} else {
		    issueWarning $ob_id $msg
		}
	    } elseif { $earliestStartTime <= $currentTime } {
		# - If the minimum date/time is in the past (before now)
		set msg "The earliest start date/time defined in the absolute time interval(s) has already passed."
		issueWarning $ob_id $msg
	    }
	}

	# [General Check] Check target for Zenith avoidance and airmass constraint:
	set targetGoesCloseToZenith 0
	if { $obRunMode == "s" && $targetOK } {
	    # Target info
	    set targDelta [expr $OBdata($ob_id,dec_marcsec) / 3600000.]
	    set csAirmass $OBdata($ob_id,csAirmass)

	    #the airmass at z=5 is (1./cos(5./$DegreePerRadian)) = 1.00381983748
	    if { $csAirmass < 1.004 } {
		issueError $ob_id "Airmass cannot be set to $csAirmass." $runID
	    } else {
		#for airmass >= 1.004

		# Latitude of VLT Observatory (Paranal = -24:37:30)
		set latitude -24.63

		# Conversion factor rad => deg
		set DegreePerRadian 57.29578

		# Value of PI
		set PI  3.14159

		# Get sin(latitude) * sin(targDelta/3600.)
		set AA [expr sin($latitude/$DegreePerRadian) * sin($targDelta/$DegreePerRadian)]

		#Get cos(latitude) * cos(targDelta/3600.)
		set BB [expr cos($latitude/$DegreePerRadian) * cos($targDelta/$DegreePerRadian)]

		#Get the equivalent altitude (in radians) for the airmass constraint sin(asin(1/airmass)) = 1/airmass
		set altX_csAirmass [expr 1.0/$csAirmass]

		if { -26.6 <= $targDelta && $targDelta <= -22.6 } {
		    #This is for declinations within zenith-avoidence zone:

		    #Get the hour angle of the target at the zenith avoidance altitude of 85deg sin(85deg)
		    set altX_zenithAvoidance [expr sin(85./$DegreePerRadian) ]

		    set HAz [expr (24./360.) * acos( ( $altX_zenithAvoidance - $AA) / $BB ) * $DegreePerRadian]

		    #Get the hour angle of the target when the target rises above the altitude of the airmas constraint
		    set HAa [expr (24./360.) * acos( ( $altX_csAirmass - $AA) / $BB ) * $DegreePerRadian]

		    set exec_hours [expr $execTime / 3600. ]
		    if { ($HAa - $HAz) < $exec_hours } {
			set printedTime [formatPrintedTime $execTime]
			set pHAa [formatPrintedTime [expr $HAa * 3600]]
			set pHAz [formatPrintedTime [expr $HAz * 3600]]
			set msg "The execution time of $printedTime is greater than the time available for the observation.\n     HA at which target rises above the airmass constraint ($csAirmass): $pHAa \n     HA at which target reaches zenith avoidance: $pHAz "
			issueError $ob_id $msg $runID
		    } elseif { $csAirmass <= 1.5 } {
			issueWarning $ob_id "The target crosses the VISTA zenith target avoidance region.\n         You may wish to review your airmass constraint to increase observability."
			set targetGoesCloseToZenith 1
		    }
		} else {
		    #This is for declinations outside zenith-avoidence zone:

		    set dec [expr $OBdata($ob_id,dec_marcsec) / 3600000. ]
		    set zenthDistance [expr abs($latitude - $dec)]
		    #get min airmass
		    set targAirmassMin [expr (1. / cos($zenthDistance/ $DegreePerRadian) ) ]
		    #get max Airmass
		    set zenthDistanceMax [expr (abs(-90 - $dec) * 2 + $zenthDistance)]
		    #this is needed for circumpolar stars
		    set targAirmassMax [expr (1. / cos($zenthDistanceMax/ $DegreePerRadian) ) ]

		    # only if the targAirmassMax > csAirmass i.e. the target gets below the airmass constraint at some point and can rise above it
		    # do the following check; Especially required for circumpolar objects. Note $targAirmassMax is -ve for +dec objects

		    set refDec [expr -90 - $dec]

		    if { $dec >= $refDec || ( $dec < $refDec && $targAirmassMax > $csAirmass)  } {

			#Get the hour angle of the target when the target rises above the altitude of the airmas constraint
			set HAa [expr (24./360.) * acos( ( $altX_csAirmass - $AA) / $BB ) * $DegreePerRadian]

			set exec_hours [expr $execTime / 3600. ]
			set tavail [expr $HAa * 2]
			set exec_hours [expr $execTime / 3600. ]
			set printedTime [formatPrintedTime $execTime]
			set pHAa [formatPrintedTime [expr $HAa * 3600]]
			set ptA [formatPrintedTime [expr $tavail * 3600]]

			if { $tavail < $exec_hours } {
			    set printedTime [formatPrintedTime $execTime]
			    set pHAa [formatPrintedTime [expr $HAa * 3600]]
			    set ptA [formatPrintedTime [expr $tavail * 3600]]
			    set msg "The execution time of $printedTime is greater than the time available for the observation, given your airmass constraint.\n     HA at which target rises above the airmass of $csAirmass : $pHAa \n     Time available for observation: $ptA "
			    issueError $ob_id $msg $runID
			}
		    }
		} ; #End else - dec not in zenith avoidence zone
	    } ; #End the airmass > at z=5
	} ;# End Check target for Zenith avoidance and airmass constraint

	# [General Check] Check epoch in range 1950.0-2050.0
	checkEpoch $ob_id $runID

	# [General Check] check on proper motions Exceptions for: VIRCAM: ERROR (SM and VM)
	# Proper motions are expressed in arcsec/year
	# in the target package
	set pma $OBdata($ob_id,pm_ra)
	set pmd $OBdata($ob_id,pm_dec)

	if { [max [expr abs($pma)] [expr abs($pmd)]] > 0.0 } {

	    set msg "Please set the proper motions to 0.0"
	    issueError $ob_id $msg $runID

	}

	# [General Check] TOO/RRM in container check
	if { $obRunMode == "s" } {
	    if { $Phase1data(tooFlag) == 1 && $isContainer } {

		if { [regexp -nocase {^(TOO_)} $OBdata($ob_id,name)]  && $containerType == "TIME_LINK"} {
		    set msg "TOO OBs should not be placed in a Time link."
		    issueError $ob_id $msg $runID
		}
		if { [regexp -nocase {^(RRM_)} $OBdata($ob_id,name)] } {
		    set msg "RRM OBs should not belong in a Group, Concatenation, or Time link. "
		    issueError $ob_id $msg $runID
		}
	    }
	}

	# PWV Constraint [General Check]. Value for default must be the same as in the constraints xsd files
	checkPWV $ob_id 30

	#PLEASE NOTE: THIS SECTION OF THE EVM CHECKS IS NOT PERFORMED FOR PUBLIC SURVEY RUNS.
	# isPublicSurvey in common/library checks for 7920 and 9820 currently)

	if { $Phase1data(isPublicSurvey) != 1 } {

	    #######################################
	    # Phase 2 vs. Phase 1 compliance checks (SM only)
	    #######################################

	    if { $obRunMode == "s" } {

		# Check expected image quality
		checkImageQuality $objRef $runID

		# Transparency constraint

		if { $OBdata($ob_id,csTransp) == "1PHO" && $Phase1data(skyTransp) == "CLR" } {

		    set msg "Phase 2 constraints must closely follow what was requested in the Phase 1 proposal.\n         The allowed Sky Transparency constraint for this OB is 'Clear', 'Thin' or 'Thick'."
		    issueWarning $ob_id $msg

		} elseif { $obRunMode == "s" && ($OBdata($ob_id,csTransp) == "1PHO" || $OBdata($ob_id,csTransp) == "2CLR") \
			&& $Phase1data(skyTransp) == "THN" } {

		    set msg "Phase 2 constraints must closely follow what was requested in the Phase 1 proposal.\n         The allowed Sky Transparency constraint for this OB is 'Thin' or 'Thick'."
		    issueError $ob_id $msg $runID

		} elseif { ($OBdata($ob_id,csTransp) == "2CLR" || $OBdata($ob_id,csTransp) == "3THN" || $OBdata($ob_id,csTransp) == "4THK") \
			&& $Phase1data(skyTransp) == "PHO" } {

		    set msg "The Sky Transparency constraint for this OB has been relaxed with respect to that in Phase 1 ($Phase1data(skyTransp)). Is this intentional?"
		    issueInfo $ob_id $msg

		} elseif { ($OBdata($ob_id,csTransp) == "3THN" || $OBdata($ob_id,csTransp) == "4THK") && \
			$Phase1data(skyTransp) == "CLR" } {

		    set msg "The Sky Transparency constraint for this OB has been relaxed with respect to that in Phase 1 ($Phase1data(skyTransp)). Is this intentional?"
		    issueInfo $ob_id $msg

		} elseif { $OBdata($ob_id,csTransp) == "4THK" && $Phase1data(skyTransp) == "THN" } {

		    set msg "The Sky Transparency constraint for this OB has been relaxed with respect to that in Phase 1 ($Phase1data(skyTransp)). Is this intentional?"
		    issueInfo $ob_id $msg

		}

		# Moon FLI constraint

		if { $obRunMode == "s" && $OBdata($ob_id,csFLI) < 0.4 && $Phase1data(moonFLI) == "g" } {

		    set msg "Phase 2 constraints must closely follow what was requested in the Phase 1 proposal.\n            The allowed Lunar Illumination value for this OB is: 0.4 <= FLI <= 0.7 (grey) or FLI > 0.7 (bright)."
		    issueError $ob_id $msg $runID

		} elseif { $OBdata($ob_id,csFLI) < 0.4 && $Phase1data(moonFLI) == "n" } {

		    set msg "Phase 2 constraints must closely follow what was requested in the Phase 1 proposal.\n         The allowed Lunar Illumination value for this OB is: FLI > 0.7"
		    issueWarning $ob_id $msg

		} elseif { $obRunMode == "s" && $OBdata($ob_id,csFLI) >= 0.4 && $OBdata($ob_id,csFLI) <= 0.7 && $Phase1data(moonFLI) == "n" } {

		    set msg "Phase 2 constraints must closely follow what was requested in the Phase 1 proposal.\n            The allowed Lunar Illumination value for this OB is: FLI > 0.7"
		    issueError $ob_id $msg $runID

		} elseif { $OBdata($ob_id,csFLI) >= 0.4 && $Phase1data(moonFLI) == "d" } {

		    set msg "The Lunar Illumination constraint for this OB has been relaxed with respect to that in Phase 1. Is this intentional?"
		    issueInfo $ob_id $msg

		} elseif { $OBdata($ob_id,csFLI) > 0.7 && $Phase1data(moonFLI) == "g" } {

		    set msg "The Lunar Illumination constraint for this OB has been relaxed with respect to that in Phase 1. Is this intentional?"
		    issueInfo $ob_id $msg

		}

		# Checks on Phase 1 targets

		# Find closest Phase 1 target
		set closest_ti 0
		set closest_dist 1e99

		for { set ti 1 } { $ti <= $Phase1data(target,nbTargets) } { incr ti } {

		    # Compute target distance
		    set distance [computeDistance \
			$Phase1data(target,$ti,RA_arcsec) [expr $OBdata($ob_id,ra_marcsec)/1000.] \
			$Phase1data(target,$ti,DEC_arcsec) [expr $OBdata($ob_id,dec_marcsec)/1000.]  ]

		    if { $distance < $closest_dist } {
			set closest_dist $distance
			set closest_ti $ti
		    }
		}

		# Set max allowed distance to original target in arcsec
		set targetRadius 7200

		if { $Phase1data(tooFlag) != 1 && \
		    [regexp -nocase {^(MOV_|CAL_|PSF_)} $OBdata($ob_id,name)] == 0 && \
			$closest_dist > $targetRadius } {

		    set msg "The target seems to be missing from the target list in the proposal."
		    issueWarning $ob_id $msg

		}

		if { $Phase1data(tooFlag) != 1 && \
		    [regexp -nocase {^(MOV_|CAL_|PSF_)} $OBdata($ob_id,name)] == 0 && \
			$closest_dist <= $targetRadius && \
		    [string first $runLetter $Phase1data(target,$closest_ti,runRefs)] == -1} {

		    set msg "This target seems to be missing from the target list for this run of the proposal."
		    issueWarning $ob_id $msg

		}
	    } ; #SM Only
	    # End of Phase2/Phase1 compliance checks

	    # Checks on LST constraints [General Check]
	    #  OB duration must be less than LST range(s)

	    for {set ti 1} {$ti <= $OBdata($ob_id,LST,nbIntervals)} {incr ti} {

		set LSTduration [expr $OBdata($ob_id,LST,$ti,end) - $OBdata($ob_id,LST,$ti,start)] ; # seconds

		if { $LSTduration < $execTime } {
		    set msg "One or more LST ranges are shorter than the execution time of the OB."
		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    } else {
			issueWarning $ob_id $msg
		    }
		    break
		}
	    } ; #end for

	}; # End is Public Survey

	# Guide star [General Check] check that this keyword is set to CATALOGUE or SETUPFILE
	# except for OBs with these templates:
	set tplList { VIRCAM_img_acq_tile VIRCAM_img_acq_quick VIRCAM_img_acq_random }

	if { [lsearch -exact $tplList $OBdata($ob_id,0,tplName)] == -1 } {
	    checkGuidestar  $ob_id
	}

	# Check on FC metadata [General Check]
	checkFC_Metadata $ob_id $objRef

    }; # end of loop on OBs

    # ^^ End Checks above for OB-level only

    # ----------------------
    # CONTAINER-LEVEL CHECKS
    # ----------------------

    # [General Check] Check the CONSISTENCY OF THE WEATHER CONSTRAINTS within a concatenation
    if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && \
	    $objSelected } {
	if { [llength [uniq $concat_seeing_list]] > 1 } {
	    issueWarning $containerID "$containerName: The seeing constraint is not the same for all OBs in the container, \n please check its influence on the science goal."
	}
	if { [llength [uniq $concat_fli_list]] > 1 } {
	    issueWarning $containerID "$containerName: The FLI is not the same for all OBs in the container, \n please check its influence on the science goal."
	}
	if { [llength [uniq $concat_transp_list]] > 1 } {
	    issueWarning $containerID "$containerName: The transparency is not the same for all OBs in the container, \n please check its influence on the science goal."
	}
	if { [llength [uniq $concat_twilight_list]] > 1 } {
	    issueWarning $containerID "$containerName: The twilight is not the same for all OBs in the container, \n please check its influence on the science goal."
	}
    }

    # [General Check] Compute RA Range of targets within concatenation to ensure feasibility
    if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && \
	    $objSelected } {

	set lim75 [format "%.2f" [expr 75 * 3600]]
	set lim45 [format "%.2f" [expr 45 * 3600]]
	set maxSep [angularSeparationCheck $ra_dec_list]

	if { $maxSep > $lim75 } {

	    set msg "Due to the large angular separation between the targets (>75 deg), the scheduling of such a \n container is extremely difficult."
	    issueErrorWaiver $containerID "$containerName: $msg" $runID 9

	} elseif { $maxSep > $lim45 } {

	    set msg "Due to the large angular separation between the targets (>45 deg), the scheduling of such a \n container is particularly difficult."
	    issueWarning $containerID "$containerName: $msg"
	}
    }

    #  [General Check] Total duration of concatenation must not exceed 65 minutes
    if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && $objSelected && \
	    $totalTimeConcat > 3900 && $obRunMode == "s" } {

	set printedTime [formatPrintedTime $totalTimeConcat]

	set msg "$containerName: Execution time for this concatenation exceeds 1h05m (found $printedTime)"
	issueErrorWaiver $containerID $msg $runID 2
    }

    #  [General Check] Absolute time windows of a Concatenation
    if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && \
	    $objSelected } {
	if { [llength [uniq $concat_OB_absTI_list]] > 1 } {
	    set msg "$containerName: OBs that have absolute time intervals defined must have the same time windows."
	    issueError $containerID $msg $runID
	}
    }

    #  LST in a Concatenation [General Check]
    if { $obRunMode == "s" && $isContainer && $containerType == "CONCATENATION" && \
	    $objSelected && $hasLSTinConcat } {
	set msg ""
	#puts "concat_LST_list $concat_LST_list [llength $concat_LST_list]"
	#determine each of the continuous intervals present in the set of LST intervals
	set continuousLSTlist {}
	set numOBs [llength $concat_LST_list]
	if { $numOBs > 0 } {
	    #add intervals of the first OB
	    set continuousLSTlist [lindex [lindex $concat_LST_list 0] 1]
	    #puts "continuousLSTlist $continuousLSTlist"
	    # define newViableForNextOB which will contain the list of defined intervals that the next OB will be checked against, start with the intervals of the first OB
	    # add all intervals to the list of continuous LST intervals starting with the first
	    set newViableForNextOB $continuousLSTlist

	    # for each OB (starting with the second OB) in the concatenation
	    for {set k 1} {$k < $numOBs} {incr k} {
		#get list of intervals for the OB
		set ob_LSTlist [lindex [lindex $concat_LST_list $k] 1]
		#for each interval in the continuousLSTlist, see if you need to change the bounds and then get a new list of viable intervals
		# now the new list of viable intervals becomes the continuousLSTlist before going to the next OB
		set continuousLSTlist $newViableForNextOB

		#puts "*** starting OB #[expr $k + 1] check against $continuousLSTlist"
		# reset newViableForNextOB
		set newViableForNextOB {}
		#also define newContinuousList which will contain the overlapping LST intervals which is used at the end of the concatenation
		set newContinuousList {}
		for {set jj 0} {$jj < [llength $continuousLSTlist]} {incr jj} {
		    for {set ii 0} {$ii < [llength $ob_LSTlist]} {incr ii} {
			set newInterval [unionInt [lindex [lindex $continuousLSTlist $jj] 0] [lindex [lindex $continuousLSTlist $jj] 1] \
			    [lindex [lindex $ob_LSTlist $ii] 0] [lindex [lindex $ob_LSTlist $ii] 1] ]
			#puts "newInterval $newInterval"
			if { $newInterval != [list 0 0] } {
			    lappend newContinuousList $newInterval
			    #puts "newContinuousList $newContinuousList"
			    lappend newViableForNextOB [lindex $ob_LSTlist $ii]
			}
		    } ; # end list of time intervals in the OB
		} ; # end list of time intervals in the continuousLSTlist

		if { [llength $newViableForNextOB] == 0 } {
		    set msg "$containerName: OB #[expr $k + 1] in the concatenation does not have LST intervals that overlap with OB #$k.\n        Overlap is necessary for OBs to be executed back-to-back."
		    break
		}

		# when no next OB, then newContinuousList is the continuousLSTlist else newViableForNextOB is the continuousLSTlist (beginning of the loop)
		set continuousLSTlist $newContinuousList
		#puts "viable intervals for the next OB $newViableForNextOB length [llength $newViableForNextOB]"
	    } ; # END for each OB

	    if { $msg != "" } {
		issueError $containerID $msg $runID
	    } else {
		#Now check for duration of LST
		#puts "continuousLSTlist $continuousLSTlist"
		for {set tj 0} {$tj < [llength $continuousLSTlist]} {incr tj} {
		    set startLST [lindex [lindex $continuousLSTlist $tj] 0]
		    set endLST  [lindex [lindex $continuousLSTlist $tj] 1]

		    set LSTduration [expr ($endLST - $startLST)  * 3600]
		    #puts "startLST $startLST endLST $endLST ---- LST duration [formatPrintedTime $LSTduration]"
		    if { $LSTduration < $totalTimeConcat } {
			set printedTime [formatPrintedTime $totalTimeConcat]
			set msg "$containerName: The LST interval(s) available to execute the concatenation in should be greater than the execution time of $printedTime"
			issueError $containerID $msg $runID
			break
		    }
		}
	    }
	} ; # end if num of OBs > 0

    } ; # end LST in concat check

} ; #End proc checkVIRCAMLevel1

#### VIRCAM Level 2 Checks - general OB structure

proc checkVIRCAMLevel2 { object } {

    global evmVIRCAMTpl
    global OBdata Phase1data

    set templateListInstrument [getArrayElements evmVIRCAMTpl]

    # Get 'object type' (first element of object list)
    set objType [lindex $object 0]

    set listOfObsBlocks {}
    set isContainer false

    # If the object type is a list, then we are dealing
    # with a container: the actual type has to be fetched
    # deeper inside the object list
    if { [llength $objType] > 1 } {

	set containerInfo [lindex $object 0]

	set objType [lindex $containerInfo 0] ; # Type of container
	set containerID [lindex $containerInfo 1] ; # Container ID
	set objRef  [lindex $containerInfo 2] ; # Object reference (Java handle)
	set objSelected [lindex $containerInfo 3] ; # Is container selected or not
	set containerName [evmName $objRef]

	set listOfObsBlocks [lindex $object 1]
	set isContainer true

    } else {

	set ob_id [lindex $object 1] ; # OB ID
	set objRef [lindex $object 2] ; # Object reference (Java handle)
	set objSelected [lindex $object 3] ; # Is object part of selection or not
	set obName $OBdata($ob_id,name)

	set listOfObsBlocks [list $object]

    }

    # Determine observing mode using first OB
    set ob [lindex $listOfObsBlocks 0]
    set ob_id [lindex $ob 1]
    set obRunMode $OBdata($ob_id,runMode)
    set runID $OBdata($ob_id,runID)

    #
    # Loop on all OBs
    #

    set obIdx 0

    foreach ob $listOfObsBlocks {

	incr obIdx

	set isOBSelected [lindex $ob 3]

	set ob_id [lindex $ob 1]

	# vv SETUP FOR LATER CONTAINER-LEVEL CHECKS vv

	####### nothing to setup #########

	# ^^ END OF SETUP FOR LATER CONTAINER-LEVEL CHECKS ^^

	# Is the OB actually part of the selection? If not, skip it..
	if { !$isOBSelected } { continue }

	# vv Checks below are for OB-level only in this loop vv

	# DETERMINE: obType (Observation/Calibration)
	#
	set obType $OBdata($ob_id,type)

	#
	# If OB = Calib, we skip the rest of the tests (irrelevant)
	#
	if { $obType == "C" } {
	    set msg "Calibration OB: no check performed"
	    issueInfo $ob_id $msg
	    continue
	}

	# Get 'mode' of acquisition
	set acqName $OBdata($ob_id,0,tplName)
	regexp {^[^,]*} [getArrayIndex evmVIRCAMTpl $acqName] acqMode

	set typePattern ""

	# Loop on all templates
	for { set i 0 } { $i < $OBdata($ob_id,nbTpl) } { incr i } {

	    set tplName $OBdata($ob_id,$i,tplName)

	    #If the Phase 1 TOO flag is OFF, then:
	    #use the or condition which can then be used in test OBs where the Phase1data is not correct

	    if { $Phase1data(tooFlag) != 1 } {

		regexp {^[^,]*} [getArrayIndex evmVIRCAMTpl $tplName] tplMode
		# Check consistency of template with acquisition ERROR (SM)
		if { $obRunMode == "s" && $tplMode != "" && $acqMode != "" && $tplMode != $acqMode } {

		    set msg "$tplName (\#$i): cannot be used with the chosen acquisition template."
		    issueError $ob_id $msg $runID

		} elseif { $tplName == "VIRCAM_img_obs_paw" } {

		    if { $obRunMode == "s" } {
			set msg "$tplName (\#$i): It is recommended that you use VIRCAM_img_obs_tileX science templates.\nUse of VIRCAM_img_obs_paw science template requires a pre-approved waiver."
			issueErrorWaiver $ob_id $msg $runID 6
		    } else {
			set msg "$tplName (\#$i): It is recommended that you use VIRCAM_img_obs_tileX science templates."
			issueWarning $ob_id $msg
		    }

		} elseif { [lsearch -exact $templateListInstrument $tplName] == -1 } {

		    set msg "$tplName (\#$i): this template should not be used for regular observations!"
		    issueWarning $ob_id $msg

		}

	    } else {

		#If the Phase 1 TOO flag is ON or the OBname begins with a TOO/RRM, then:
		set acq_tile_tmpls { "VIRCAM_img_acq_tile" "VIRCAM_img_obs_tile1"   "VIRCAM_img_obs_tile3"   "VIRCAM_img_obs_tile6"   "VIRCAM_img_obs_tile6sky" }
		set acq_quick_tmpls { "VIRCAM_img_acq_quick" "VIRCAM_img_obs_paw" }
		#evmInfo "acqName: $acqName tplName: $tplName acq_quick_tmpls: $acq_quick_tmpls [lsearch -exact $acq_quick_tmpls $tplName]"
		#Check that the templates are associated correctly, else ERROR (SM)
		if { ( $acqName == "VIRCAM_img_acq_tile" && [lsearch -exact $acq_tile_tmpls $tplName] == -1 )  || \
			( $acqName == "VIRCAM_img_acq_quick" && [lsearch -exact $acq_quick_tmpls $tplName] == -1 ) } {

		    set msg "$tplName (\#$i): cannot be used with the chosen acquisition template."

		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    }
		} elseif { $acqName == "VIRCAM_img_acq_tile" && ( $tplName == "VIRCAM_img_obs_tile1" || $tplName == "VIRCAM_img_obs_tile3") } {
		    #Check that the target does not fall into a gap between detectors
		    set msg "Check that the target does not fall into a gap between detectors."
		    issueInfo $ob_id $msg

		} elseif { [lsearch -exact $acq_tile_tmpls $tplName] == -1  && [lsearch -exact $acq_quick_tmpls $tplName] == -1 } {

		    set msg "$tplName (\#$i): Template cannot be used for a TOO run."
		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    }

		}

	    }

	    if { [regexp {_(obs|cal)_} $tplName dummy tplType] } {
		append typePattern $tplType
	    }

	}

	# Check for mixed obs/cal OBs
	if { ![regexp {^(obs)+$|^(cal)+$} $typePattern] } {

	    set msg "you cannot mix 'obs' and 'cal' templates in the same OB"

	    if { $obRunMode == "s" } {
		issueError $ob_id $msg $runID
	    } else {
		issueWarning $ob_id $msg
	    }

	}

    }; # end of loop on OBs

    # ^^ End Checks above for OB-level only

    # ----------------------
    # CONTAINER-LEVEL CHECKS
    # ----------------------

    ####### no container-level checks #########

    # -----------------------------
    # END OF CONTAINER-LEVEL CHECKS
    # -----------------------------
}

#### VIRCAM Level 3 Checks - detailed checks on parameters

proc checkVIRCAMLevel3 { object } {

    global OBdata evmDITLimits Phase1data

    # Get 'object type' (first element of object list)
    set objType [lindex $object 0]

    set listOfObsBlocks {}
    set isContainer false

    # If the object type is a list, then we are dealing
    # with a container: the actual type has to be fetched
    # deeper inside the object list
    if { [llength $objType] > 1 } {

	set containerInfo [lindex $object 0]

	set objType [lindex $containerInfo 0] ; # Type of container
	set containerID [lindex $containerInfo 1] ; # Container ID
	set objRef  [lindex $containerInfo 2] ; # Object reference (Java handle)
	set objSelected [lindex $containerInfo 3] ; # Is container selected or not
	set containerName [evmName $objRef]

	set listOfObsBlocks [lindex $object 1]
	set isContainer true

    } else {

	set ob_id [lindex $object 1] ; # OB ID
	set objRef [lindex $object 2] ; # Object reference (Java handle)
	set objSelected [lindex $object 3] ; # Is object part of selection or not
	set obName $OBdata($ob_id,name)

	set listOfObsBlocks [list $object]

    }

    # Determine observing mode using first OB
    set ob [lindex $listOfObsBlocks 0]
    set ob_id [lindex $ob 1]
    set obRunMode $OBdata($ob_id,runMode)
    set runID $OBdata($ob_id,runID)

    #
    # Loop on all OBs
    #

    set obIdx 0

    foreach ob $listOfObsBlocks {

	incr obIdx

	set isOBSelected [lindex $ob 3]

	# Is the OB actually part of the selection? If not, skip it..
	if { !$isOBSelected } { continue }

	set ob_id [lindex $ob 1]

	# DETERMINE: obType (Observation/Calibration)
	#
	set obType $OBdata($ob_id,type)

	#
	# If OB = Calib, we skip the rest of the tests (irrelevant)
	#
	if { $obType == "C" } {
	    set msg "Calibration OB: no check performed"
	    issueInfo $ob_id $msg
	    continue
	}

	#
	# Checks on acquisition template
	#

	set tplList { VIRCAM_img_acq VIRCAM_img_acq_quick }

	if { [lsearch -exact $tplList $OBdata($ob_id,0,tplName)] != -1 } {

	    # Warning if 'confirm guide star' selected
	    if { $OBdata($ob_id,0,TEL.AG.CONFIRM) == "T" } {

		set msg "$OBdata($ob_id,0,tplName) (\#0): selecting 'Confirm Guide Star?' induces an additional overhead."
		issueWarning $ob_id $msg

	    }

	    # Warning if 'confirm active optics' selected
	    if { $OBdata($ob_id,0,TEL.AO.CONFIRM) == "T" } {

		set msg "$OBdata($ob_id,0,tplName) (\#0): selecting 'Confirm Active Optics?' induces an additional overhead."
		issueWarning $ob_id $msg

	    }
	}

	set tplList { VIRCAM_img_acq VIRCAM_img_acq_tile VIRCAM_img_acq_quick }

	if { [lsearch -exact $tplList $OBdata($ob_id,0,tplName)] != -1 } {

	    # Warning if seeing < 1.0 and active optics priority != HIGH
	    if { $obRunMode == "s" && $OBdata($ob_id,csSeeing) < 1.0 && \
		    $OBdata($ob_id,0,TEL.AO.PRIORITY) != "HIGH" } {

		set msg "$OBdata($ob_id,0,tplName) (\#0): seeing constraint < 1.0 but Active Optics Priority not set to HIGH."
		issueErrorWaiver $ob_id $msg $runID 4
	    }

	    # Warning if active optics priority is set to LOW
	    if { $OBdata($ob_id,0,TEL.AO.PRIORITY) == "LOW" } {

		set msg "$OBdata($ob_id,0,tplName) (\#0): Active Optics Priority is set to LOW."
		issueWarning $ob_id $msg
	    }

	}

	# Warnings if auto-guiding and/or active optics not enabled
	if { $OBdata($ob_id,0,tplName) == "VIRCAM_img_acq_quick" } {

	    if { $OBdata($ob_id,0,TEL.AG.START) == "F" } {
		set msg "$OBdata($ob_id,0,tplName) (\#0): Autoguiding is not enabled."
		issueWarning $ob_id $msg
	    }

	    if { $OBdata($ob_id,0,TEL.AO.START) == "F" } {
		set msg "$OBdata($ob_id,0,tplName) (\#0): Active Optics are not enabled."
		issueWarning $ob_id $msg
	    }

	}

	# Warning if AO priority set to HIGH but not using auto-guider nor active optics
	if { $OBdata($ob_id,0,tplName) == "VIRCAM_img_acq_quick" || \
		$OBdata($ob_id,0,tplName) == "VIRCAM_img_acq" } {

	    if { $OBdata($ob_id,0,TEL.AO.PRIORITY) == "HIGH" && \
		    ($OBdata($ob_id,0,TEL.AG.START) == "F" || $OBdata($ob_id,0,TEL.AO.START) == "F") } {

		set msg "$OBdata($ob_id,0,tplName) (\#0): Active Optics Priority is set to 'HIGH' but Autoguiding and/or Active Optics are not enabled!"

		if { $obRunMode == "s" } {
		    issueError $ob_id $msg $runID
		} else {
		    issueWarning $ob_id $msg
		}
	    }

	}

	#ToO acquisition on Det. 11 check ERROR/WAIVER(SM)
	if { $obRunMode == "s" && $Phase1data(tooFlag) == 1 && $OBdata($ob_id,0,tplName) == "VIRCAM_img_acq_quick"} {
	    if { $OBdata($ob_id,0,TEL.AG.START) != "T" || \
		    $OBdata($ob_id,0,TEL.AG.CONFIRM) != "T" || \
		    $OBdata($ob_id,0,TEL.AO.START) != "T" || \
		    $OBdata($ob_id,0,TEL.AO.CONFIRM) != "T" || \
		    $OBdata($ob_id,0,TEL.ROT.OFFANGLE) != "0.0" || \
		    $OBdata($ob_id,0,TEL.TARG.X) != "39.0" || \
		    $OBdata($ob_id,0,TEL.TARG.Y) != "-29.3" } {

		set msg  "Your TOO acquisition is non-standard: Autoguiding and active optics should be enabled,\n"
		append msg "     and guide star and active optics should be confirmed. Rotator Angle on Sky should be 0.0.\n"
		append msg "     Set X, Y coordinates of pointing to 39.0, -29.3."
		# (7) Allow Non-standard ToO acquisition on Det. 11
		issueErrorWaiver $ob_id $msg $runID 7

	    }
	}

	# Check first filter used after acquisition (should be the same)
	if { $OBdata($ob_id,type)!= "C" && \
	    [lindex $OBdata($ob_id,1,INS.FILTER.NAME) 0] != $OBdata($ob_id,0,INS.FILTER.NAME) } {

	    set msg "$OBdata($ob_id,0,tplName) (\#0): the filter used in the acquisition is not the same as the first filter in the following template (implies additional overhead)."
	    issueInfo $ob_id $msg

	}

	#
	# Loop on all templates
	#

	for { set i 0 } { $i < $OBdata($ob_id,nbTpl) } { incr i } {

	    # *******************************************************
	    # ==> CATEGORY: FILTERS, SLIT, GRISMS...
	    # *******************************************************

	    # Check if nesting = PFJME but only one filter used
	    if { $OBdata($ob_id,$i,SEQ.NESTING) == "PFJME" && \
		[llength $OBdata($ob_id,$i,INS.FILTER.NAME)] == 1 } {

		set msg "$OBdata($ob_id,$i,tplName) (\#$i): using Nesting 'PFJME' but template uses only one filter."
		issueWarning $ob_id $msg

	    }

	    # Check if using more than one filter for nestings FPJME and FJPME
	    if { $OBdata($ob_id,$i,SEQ.NESTING) != "PFJME" && \
		[llength $OBdata($ob_id,$i,INS.FILTER.NAME)] > 1 } {

		set msg "$OBdata($ob_id,$i,tplName) (\#$i): to simplify observatory operations it is recommended that only one filter be entered per template with nestings FPJME and FJPME."
		if { $obRunMode == "s" } {
		    issueError $ob_id $msg $runID
		} else {
		    issueWarning $ob_id $msg
		}
	    }

	    # Check for over-specified Moon constraints (science templates only hence i >= 1)
	    if { $OBdata($ob_id,type)!= "C" && $i >= 1 && \
		[lsearch -regexp $OBdata($ob_id,$i,INS.FILTER.NAME) "(J|H|Ks)"] != -1 && \
		    ($OBdata($ob_id,csFLI) < 0.7 || $OBdata($ob_id,csMoonDist) > 50) } {

		set msg "$OBdata($ob_id,$i,tplName) (\#$i): for the filters J, H or Ks your moon constraints may be over-specified."
		issueInfo $ob_id $msg

	    }

	    # Check for consecutive identical filters
	    set filterList $OBdata($ob_id,$i,INS.FILTER.NAME)
	    set uniqueFilterList [lsort [uniq $filterList]]

	    foreach f $uniqueFilterList {

		if { [regexp "$f $f" $filterList] } {

		    set msg "$OBdata($ob_id,$i,tplName) (\#$i): filter $f is used twice consecutively in the list of filters."
		    issueWarning $ob_id $msg

		}

	    }

	    # ***************************************************
	    # ==> CATEGORY: EXPOSURE TIME AND NUMBER OF EXPOSURES
	    # ***************************************************

	    # Check number of DITs vs. NDITs
	    if { [llength $OBdata($ob_id,$i,DET1.DIT)] != [llength $OBdata($ob_id,$i,DET1.NDIT)] } {

		set msg "$OBdata($ob_id,$i,tplName) (\#$i): the numbers of DITs and NDITs should be the same."
		issueWarning $ob_id $msg

	    }

	    # Check number of DITs vs. number of filters
	    if { $OBdata($ob_id,$i,DET1.DIT) != "" && $OBdata($ob_id,$i,INS.FILTER.NAME) != "" } {

		set nbDit [llength $OBdata($ob_id,$i,DET1.DIT)]
		set nbFilt [llength $OBdata($ob_id,$i,INS.FILTER.NAME)]

		if { $nbDit > 1 && ($nbDit != $nbFilt) } {

		    set msg "$OBdata($ob_id,$i,tplName) (\#$i): the number of DITs does not correspond to the number of filters.\n         "
		    append msg "Number of DITs (and NDITs) should either be equal to one or to the number of filters.\n         "
		    append msg "Remark: recommend having N templates with one filter rather than one template with N filters."

		    issueWarning $ob_id $msg

		}

	    }

	    # Minimum time between exposures
	    set tplList { "VIRCAM_img_obs_tile1" "VIRCAM_img_obs_tile3" \
			      "VIRCAM_img_obs_tile6" "VIRCAM_img_obs_tile6sky" "VIRCAM_img_obs_paw" }

	    if { $obRunMode == "s" && [lsearch -exact $tplList $OBdata($ob_id,$i,tplName)] != -1 } {

		if { [set nbDit [llength $OBdata($ob_id,$i,DET1.DIT)]] > [set nbNdit [llength $OBdata($ob_id,$i,DET1.NDIT)]] } {

		    set ditList $OBdata($ob_id,$i,DET1.DIT)
		    set nDitList $OBdata($ob_id,$i,DET1.NDIT)
		    set filterList $OBdata($ob_id,$i,INS.FILTER.NAME)

		    while {$nbDit > $nbNdit} {

			set nDitList [concat $nDitList \
			    [lrange $nDitList 0 [expr $nbDit - $nbNdit - 1]]]

			set filterList [concat $filterList \
			    [lrange $filterList 0 [expr $nbDit - $nbNdit - 1]]]

			set nbNdit [llength $nDitList]
		    }

		} elseif { [set nbNdit [llength $OBdata($ob_id,$i,DET1.NDIT)]] > [set nbDit [llength $OBdata($ob_id,$i,DET1.DIT)]] } {

		    set ditList $OBdata($ob_id,$i,DET1.DIT)
		    set filterList $OBdata($ob_id,$i,INS.FILTER.NAME)

		    while {$nbNdit > $nbDit} {

			set ditList [concat $ditList \
			    [lrange $ditList 0 [expr $nbNdit - $nbDit - 1]]]
			set filterList [concat $filterList \
			    [lrange $filterList 0 [expr $nbNdit - $nbDit - 1]]]
			set nbDit [llength $ditList]
		    }

		    set nDitList $OBdata($ob_id,$i,DET1.NDIT)

		} else {

		    set ditList $OBdata($ob_id,$i,DET1.DIT)
		    set nDitList $OBdata($ob_id,$i,DET1.NDIT)
		    set filterList $OBdata($ob_id,$i,INS.FILTER.NAME)

		}

		set minTimeFlag false
		set minTimeFlag2 false
		set maxTimeFlag false
		set minTime 1e99
		set minTime2 1e99
		set maxTime 0

		foreach f [uniq $filterList] {
		    set minDit($f) 1e99
		    set maxDit($f) 0
		}

		# Determine number of exposures
		switch -- $OBdata($ob_id,$i,SEQ.NESTING) {

		    "FPJME" -
		    "PFJME" {

			switch -- $OBdata($ob_id,$i,SEQ.JITTER.ID) {

			    "Single"   { set nbJitter 1}

			    "Jitter2d" -
			    "Jitter2u" -
			    "Jitter2da" -
			    "Jitter2ua" { set nbJitter 2}

			    "Jitter3u" -
			    "Jitter3d" { set nbJitter 3}

			    "Jitter4u" { set nbJitter 4}

			    "Jitter5n" -
			    "Jitter5z" { set nbJitter 5}

			    "Jitter9s" { set nbJitter 9}

			    "Jitter25s" { set nbJitter 25}

			    "Jitter30r1" { set nbJitter 30}

			    "Random" { set nbJitter $OBdata($ob_id,$i,SEQ.JITTER.NJITTER)}

			    default   { set nbJitter 1}
			}

			set nexpo $nbJitter
		    }

		    "FJPME" {

			set nexpo 1
		    }

		    default {
			set nexpo 1
		    }

		}

		# Handle case of DIT, NDIT and Filter lists having different lengths:

		# Longest list drives the loop
		set L [max [max [llength $ditList] [llength $nDitList]] [llength $filterList]]

		# Create 'circular lists' if needed
		set cDitList $ditList
		if { [llength $ditList] < $L } {

		    set iter [expr $L/[llength $ditList] ]
		    for { set kk 1 } { $kk <= $iter } { incr kk } {
			set cDitList [concat $cDitList $ditList]
		    }
		    set cDitList [lrange $cDitList 0 [expr $L - 1]]
		}

		set cNditList $nDitList
		if { [llength $nDitList] < $L } {

		    set iter [expr $L/[llength $nDitList] ]
		    for { set kk 1 } { $kk <= $iter } { incr kk } {
			set cNditList [concat $cNditList $nDitList]
		    }
		    set cNditList [lrange $cNditList 0 [expr $L - 1]]
		}

		set cFilterList $filterList
		if { [llength $filterList] < $L } {

		    set iter [expr $L/[llength $filterList] ]
		    for { set kk 1 } { $kk <= $iter } { incr kk } {
			set cFilterList [concat $cFilterList $filterList]
		    }
		    set cFilterList [lrange $cFilterList 0 [expr $L - 1]]
		}

		for {set j 0} {$j < $L} {incr j} {

		    set dit [lindex $cDitList $j]
		    set ndit [lindex $cNditList $j]
		    set f [lindex $cFilterList $j]

		    if { $dit > $maxDit($f) } { set maxDit($f) $dit }
		    if { $dit < $minDit($f) } { set minDit($f) $dit }

		    if { $j < [expr [llength $cDitList] - 1] } {

			set filterTime [getFilterChangeTime $f \
			    [lindex $cFilterList [expr $j + 1]] ]

			if { [set tbe2 [expr $ndit * ($dit + 2) + $filterTime ]] < 45 } {

			    set minTimeFlag2 true
			    if { $tbe2 < $minTime2 } { set minTime2 $tbe2 }

			}

		    }

		    if { [set tbe [expr $ndit * ($dit + 2)]] < 45 } {

			set minTimeFlag true
			if { $tbe < $minTime } { set minTime $tbe }

		    }

		    if { [set tbe [expr $dit * $ndit]] > 300 } {

			set maxTimeFlag true
			if { $tbe > $maxTime } { set maxTime $tbe }

		    }

		}

		if { $OBdata($ob_id,$i,SEQ.NESTING) == "PFJME" && \
			$OBdata($ob_id,$i,SEQ.JITTER.ID) == "Single" && \
		    [llength [uniq $OBdata($ob_id,$i,INS.FILTER.NAME)]] > 1 && \
			$OBdata($ob_id,$i,SEQ.USTEP.ID) == "Single" } {

		    if { $minTimeFlag2 } {

			set msg "$OBdata($ob_id,$i,tplName) (\#$i): Time between subsequent moves of the telescope is shorter than 45 sec (found a minimum of [format %.0f $minTime2]sec)."
			append msg "\n         This is too short for LOWFS correction."

			if { $OBdata($ob_id,0,TEL.AO.PRIORITY) != "HIGH" } {
			    append msg "\n         Also, the AO priority is not set to HIGH so you are not getting the optimum image quality."
			}

			issueWarning $ob_id $msg

		    }

		} elseif { $minTimeFlag } {

		    set msg "$OBdata($ob_id,$i,tplName) (\#$i): (DIT+2) * NDIT should be >= 45sec to allow for LOWFS correction (found a minimum of [format %.0f $minTime]sec)."
		    append msg "\n         \[ NEXPO=$nexpo with the chosen nesting and jittering pattern \]"

		    if { $OBdata($ob_id,0,TEL.AO.PRIORITY) != "HIGH" } {
			append msg "\n         Also, the Active optics priority is not set to HIGH so you are not getting the optimum image quality."
		    }

		    issueWarning $ob_id $msg

		}

		# Overlong exposure times prevent good sky subtraction
		if { $OBdata($ob_id,$i,INS.FILTER.NAME) != "NB118" && \
			$OBdata($ob_id,$i,INS.FILTER.NAME) != "NB980" && \
			$maxTimeFlag } {

		    set msg "$OBdata($ob_id,$i,tplName) (\#$i): DIT * NDIT is very long (found [format %.0f $maxTime]sec), good sky subtraction may be impossible."
		    issueWarning $ob_id $msg

		}

		# Check DITs for filters

		foreach filter [uniq $OBdata($ob_id,$i,INS.FILTER.NAME)] {

		    if { $minDit($filter) < 1.0011 } {

			set msg "$OBdata($ob_id,$i,tplName) (\#$i): the smallest DIT used with filter $filter is below the minimum possible value. The minimum value (1.0011 s) will be used instead."
			issueWarning $ob_id $msg
			set minDit($filter) 1.0011
		    }

		    if { [info exists evmDITLimits(RONlimited,$filter)] && \
			    $minDit($filter) < $evmDITLimits(RONlimited,$filter) } {

			set msg "$OBdata($ob_id,$i,tplName) (\#$i): the shortest DIT used with filter $filter ($minDit($filter) s) is such that the observations may be limited by the read-out noise.\n         "
			append msg "With the $filter filter, try using DIT >= $evmDITLimits(RONlimited,$filter) s"
			issueWarning $ob_id $msg

		    }

		    if { [info exists evmDITLimits(saturate,$filter)] && \
			    $maxDit($filter) > $evmDITLimits(saturate,$filter) } {

			set msg "$OBdata($ob_id,$i,tplName) (\#$i): the longest DIT used with filter $filter ($maxDit($filter) s) is such that the sky may reach exposure levels above 10000adu.\n         "
			append msg "With the $filter filter, try using DIT <= $evmDITLimits(saturate,$filter) s"
			issueWarning $ob_id $msg

		    }

		}

	    }

	    # *******************************
	    # ==> CATEGORY: OFFSET PARAMETERS
	    # *******************************

	    # Check jitter parameters
	    set tplList { "VIRCAM_img_obs_tile1" "VIRCAM_img_obs_tile3" "VIRCAM_img_obs_tile6" "VIRCAM_img_obs_tile6sky" }

	    if { [lsearch -exact $tplList $OBdata($ob_id,$i,tplName)] != -1 } {

		if { $OBdata($ob_id,$i,SEQ.JITTER.ID) == "Random" && \
			$OBdata($ob_id,$i,SEQ.JITTER.MAX) < 10 } {

		    set msg "$OBdata($ob_id,$i,tplName) (\#$i): Random jittering with max. jitter size less than 10 arcsec. This may result in poor sky subtraction."
		    issueWarning $ob_id $msg

		}

		if { $OBdata($ob_id,$i,SEQ.JITTER.ID) == "Single" } {

		    set msg "$OBdata($ob_id,$i,tplName) (\#$i): Jitter pattern is 'Single'. Consider whether to use jittering to improve sky subtraction."
		    issueWarning $ob_id $msg
		}
	    }

	    # Maximum sky offset for tile6sky ERROR (SM and VM)
	    if { $OBdata($ob_id,$i,tplName) == "VIRCAM_img_obs_tile6sky" } {

		if { $OBdata($ob_id,$i,SEQ.SKYOFFSET.ALPHA) == "" || \
			$OBdata($ob_id,$i,SEQ.SKYOFFSET.DELTA) == "" } {
		    set msg "$OBdata($ob_id,$i,tplName) (\#$i): you must set both RA and DEC for the sky offset."
		    issueError $ob_id $msg $runID
		} else {
		    set targetRA  [expr $OBdata($ob_id,ra_marcsec) / 1000.]
		    set targetDEC [expr $OBdata($ob_id,dec_marcsec) / 1000.]

		    regsub -all {(^\ )|(:)|(0+$)} $OBdata($ob_id,$i,SEQ.SKYOFFSET.ALPHA) "" skyAlpha
		    set skyRA [ra2arcsec $skyAlpha]
		    regsub -all {(^\ )|(:)|(0+$)} $OBdata($ob_id,$i,SEQ.SKYOFFSET.DELTA) "" skyDelta
		    set skyDEC [dec2arcsec $skyDelta]

		    set offset [computeDistance $targetRA $skyRA $targetDEC $skyDEC]

		    # If sky offset larger than 18000 arcseconds = 5 degrees, issue a WARNING
		    if { $offset > 18000 } {
			set msg "$OBdata($ob_id,$i,tplName) (\#$i): distance of the sky position from the target is larger than 5 degrees."
			issueWarning $ob_id $msg
		    } elseif { $offset < 1800 } {
			set msg "$OBdata($ob_id,$i,tplName) (\#$i): the chosen sky field is very close to the target (< 0.5 degrees)."
			issueWarning $ob_id $msg
		    }
		}
	    }

	    # *******************************
	    # ==> CATEGORY: OTHER PARAMETERS
	    # *******************************

	    if { $OBdata($ob_id,$i,tplName) == "VIRCAM_img_obs_tile6sky" && \
		    $OBdata($ob_id,$i,OCS.EXTENDED) != "FIXEDSKYOFFSET" } {

		set msg "$OBdata($ob_id,$i,tplName) (\#$i): you must select 'FIXEDSKYOFFSET' for the parameter 'Is object extended?' for the template to actually perform the sky exposures."
		issueWarning $ob_id $msg

	    }

	}; # end of loop on templates

    }; # end of loop on OBs

}

# VIRCAM Level 4 checks - Attached PAF files
proc checkVIRCAMLevel4 { object } {

    global OBdata

    # Version of SADT + matching IP version
    set SADTVersion "5.1"
    set IPVersion "109"
    set pVersion "109"

    # Get 'object type' (first element of object list)
    set objType [lindex $object 0]

    set listOfObsBlocks {}
    set isContainer false

    # If the object type is a list, then we are dealing
    # with a container: the actual type has to be fetched
    # deeper inside the object list
    if { [llength $objType] > 1 } {

	set containerInfo [lindex $object 0]

	set objType [lindex $containerInfo 0] ; # Type of container
	set containerID [lindex $containerInfo 1] ; # Container ID
	set objRef  [lindex $containerInfo 2] ; # Object reference (Java handle)
	set objSelected [lindex $containerInfo 3] ; # Is container selected or not
	set containerName [evmName $objRef]

	set listOfObsBlocks [lindex $object 1]
	set isContainer true

    } else {

	set ob_id [lindex $object 1] ; # OB ID
	set objRef [lindex $object 2] ; # Object reference (Java handle)
	set objSelected [lindex $object 3] ; # Is object part of selection or not
	set obName $OBdata($ob_id,name)

	set listOfObsBlocks [list $object]

    }

    # Determine observing mode using first OB
    set ob [lindex $listOfObsBlocks 0]
    set ob_id [lindex $ob 1]
    set obRunMode $OBdata($ob_id,runMode)
    set runID $OBdata($ob_id,runID)

    #
    # Loop on all OBs
    #

    set obIdx 0

    foreach ob $listOfObsBlocks {

	incr obIdx

	set isOBSelected [lindex $ob 3]

	# Is the OB actually part of the selection? If not, skip it..
	if { !$isOBSelected } { continue }

	set ob_id [lindex $ob 1]
	set objType [lindex $ob 0]
	set objRef [lindex $ob 2]

	# DETERMINE: obType (Observation/Calibration)
	#
	set obType $OBdata($ob_id,type)

	#
	# If OB = Calib, we skip the rest of the tests (irrelevant)
	#
	if { $obType == "C" } {
	    set msg "Calibration OB: no check performed"
	    issueInfo $ob_id $msg
	    continue
	}

	set tplList { "VIRCAM_img_obs_tile1" "VIRCAM_img_obs_tile3" "VIRCAM_img_obs_tile6" "VIRCAM_img_obs_tile6sky"}

	#
	# Loop on all templates
	#

	for { set i 0 } { $i < $OBdata($ob_id,nbTpl) } { incr i } {

	    if { [lsearch -exact $tplList $OBdata($ob_id,$i,tplName)] == -1 } { continue }

	    # Retrieve parameters associated to setup files (SEQ.REF.FILE*)
	    set paramList { "SEQ.REF.FILE1" "SEQ.REF.FILE2" \
				"SEQ.REF.FILE3" "SEQ.REF.FILE4" \
				"SEQ.REF.FILE5" "SEQ.REF.FILE6" }

	    set mandatoryKeywords { TEL.GS1.ALPHA TEL.GS1.DELTA TEL.GS1.MAG OCS.SADT.GS1.ID \
					TEL.AOSA1.ALPHA TEL.AOSA1.DELTA TEL.AOSA1.MAG OCS.SADT.AOSA1.ID \
					TEL.AOSB1.ALPHA TEL.AOSB1.DELTA TEL.AOSB1.MAG OCS.SADT.AOSB1.ID \
					OCS.SADT.TILE.RA OCS.SADT.TILE.DEC OCS.SADT.TILE.OFFANGLE OCS.SADT.PATTERN }

	    set pafFileOK_SADT 0
	    set pafFileOK_IP 0
	    set pafFileSADTVersion 0
	    set pafFileIPVersion 0

	    foreach param $paramList {

		regexp {[0-9]$} $param paramIndex

		if { $OBdata($ob_id,$i,$param) == "" } { continue }

		# Test validity of PAF file: must start with 'PAF.HDR.START'
		if { ![regexp {^PAF\.HDR\.START} [string trimleft $OBdata($ob_id,$i,$param)] ] } {

		    set msg "$OBdata($ob_id,$i,tplName) (\#$i): the attached file does not look like a regular PAF file. Please use the SADT import feature to attach the PAF file."
		    issueError $ob_id $msg $runID
		    continue

		}

		set pafTable [evmParsePAF $OBdata($ob_id,$i,$param)]

		set creatorName [evmGetPAFParameter $pafTable "PAF.CRTE.NAME"]

		set sadtused ""
		set ip ""

		regexp {SADT v([0-9\.]+) \(VIRCAM-(([0-9]+)\.[0-9][0-9])\)} $creatorName dummy sadtused full_ip ip

		set pafFileOK true

		if { $sadtused == "" } {

		    set pafFileOK false
		    incr pafFileOK_SADT

		} elseif { $sadtused <  $SADTVersion } {

		    incr pafFileSADTVersion
		}

		if { $ip == "" } {

		    set pafFileOK false
		    incr pafFileOK_IP

		} elseif { $ip != $IPVersion } {

		    incr pafFileIPVersion
		}

		if { $pafFileOK } {

		    set keywordsMissing 0
		    set keywordsMissingList {}

		    foreach key $mandatoryKeywords {

			set value [evmGetPAFParameter $pafTable $key]

			if { $value == "" } {

			    incr keywordsMissing
			    lappend keywordsMissingList $key
			    #break
			}

		    }

		    if { $keywordsMissing > 0 } {

			set msg "$OBdata($ob_id,$i,tplName) (\#$i): some keywords are missing from the setup file for pawprint $paramIndex.\n"
			append msg "         (missing keywords: $keywordsMissingList)"
			issueWarning $ob_id $msg

		    }

		    #Check that AG is always ON on the first paw-print
		    set agTplList { "VIRCAM_img_obs_tile1" "VIRCAM_img_obs_tile3" "VIRCAM_img_obs_tile6"}
		    if { [lsearch -exact $agTplList $OBdata($ob_id,$i,tplName)] != -1  && $paramIndex == 1 } {
			set tel_ag_start [evmGetPAFParameter $pafTable "TEL.AG.START"]
			if { $tel_ag_start != "T" } {
			    set msg "The use of auto-guiding in the first paw-print is mandatory."
			    issueErrorWaiver $ob_id $msg $runID 8
			}
		    }

		    if { $keywordsMissing == 0 } {

			# Check values of keywords vs. OB parameters
			set RA_PAFfile [evmGetPAFParameter $pafTable "OCS.SADT.TILE.RA"]
			set DEC_PAFfile [evmGetPAFParameter $pafTable "OCS.SADT.TILE.DEC"]

			set OFFANGLE_PAFfile [evmGetPAFParameter $pafTable "OCS.SADT.TILE.OFFANGLE"]
			set Pattern_PAFfile [evmGetPAFParameter $pafTable "OCS.SADT.PATTERN"]

			# NB: we compare the coordinates from the target package and the attached
			# PAF files. To avoid rounding issues, the comparison is done only down to
			# the 1/10th of an arcsec.

			set RA_PAFfile [format %.1f $RA_PAFfile]
			set DEC_PAFfile [format %.1f $DEC_PAFfile]

			regsub -all ":" $OBdata($ob_id,ra) "" RA_target_package
			regsub -all ":" $OBdata($ob_id,dec) "" DEC_target_package

			set RA_target_package [format %.1f $RA_target_package]
			set DEC_target_package [format %.1f $DEC_target_package]

			if { $RA_target_package != $RA_PAFfile || \
				$DEC_target_package != $DEC_PAFfile } {
			    #ERROR (SM and VM)
			    set msg "$OBdata($ob_id,$i,tplName) (\#$i): coordinates found in the attached setup file for pawprint $paramIndex differ from the ones found in the target package."
			    issueError $ob_id $msg $runID
			}

			if { [expr abs($OBdata($ob_id,0,TEL.ROT.OFFANGLE) - $OFFANGLE_PAFfile)] > 1.0 } {

			    set msg "$OBdata($ob_id,$i,tplName) (\#$i): the rotator angle found in the attached setup file for pawprint $paramIndex ($OFFANGLE_PAFfile deg) does not match the value in the acquisition template."
			    issueError $ob_id $msg $runID
			}

			if { $OBdata($ob_id,$i,SEQ.TILE.ID) != $Pattern_PAFfile } {

			    set msg "$OBdata($ob_id,$i,tplName) (\#$i): the name of the tile pattern found in the attached setup file for pawprint $paramIndex does not match the value selected in the template ($OBdata($ob_id,$i,SEQ.TILE.ID))."
			    issueError $ob_id $msg $runID
			}

			switch -- $OBdata($ob_id,$i,SEQ.JITTER.ID) {

			    "Single" { set maxJitter 0 }

			    "Jitter2d" -
			    "Jitter2u" -
			    "Jitter2da" -
			    "Jitter2ua" -
			    "Jitter3d" -
			    "Jitter3u" -
			    "Jitter5n" -
			    "Jitter30r1" { set maxJitter 10 }

			    "Jitter4u" -
			    "Jitter5z" -
			    "Jitter9s" -
			    "Jitter25s" { set maxJitter 14 }

			    "Random" {
				set maxJitter $OBdata($ob_id,$i,SEQ.JITTER.MAX)
			    }

			}

			set PAFmaxJitter [evmGetPAFParameter $pafTable "OCS.SADT.MAXJIT"]

			if { $PAFmaxJitter < [expr  $maxJitter * $OBdata($ob_id,$i,SEQ.JITTER.SCALE)] } {

			    set msg "$OBdata($ob_id,$i,tplName) (\#$i): the max jitter size found in the attached setup file for pawprint $paramIndex ($PAFmaxJitter arcsec) is incompatible with the values selected in the template."
			    issueWarning $ob_id $msg

			}

		    }

		}

	    }; # end of loop on param files ERROR (SM and VM)

	    if { $pafFileOK_SADT > 0 } {

		set msg "$OBdata($ob_id,$i,tplName) (\#$i): could not find SADT version in setup files for $pafFileOK_SADT of the pawprints. Are these proper SADT files?"
		issueError $ob_id $msg $runID
	    }

	    if { ![regexp {^[0-9]([0-9]+)\.[A-Z]-([0-9]+)} $runID dummy period proposal] } {
		regexp {^([0-9]+)\.(....)} $runID dummy period proposal
	    }

	    if { $pafFileSADTVersion > 0  } {
		set msg "$OBdata($ob_id,$i,tplName) (\#$i): the version of SADT used is $sadtused (recommended version is $SADTVersion or higher)."

		# For new runs
		if { $period >= $pVersion } {
		    #error:  the latest SADT version MUST be used:
		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    } else {
			issueWarning $ob_id $msg
		    }
		} elseif { $period == 98 && [string range $proposal 0 1] == 2 } {
		    issueError $ob_id $msg $runID
		} elseif { $period < $pVersion } {
		    issueWarning $ob_id $msg
		}

	    }

	    # ERROR (SM and VM)
	    if { $pafFileOK_IP > 0 } {

		set msg "$OBdata($ob_id,$i,tplName) (\#$i): could not find IP version in setup files for $pafFileOK_IP of the pawprints. Are these proper SADT files?"
		issueError $ob_id $msg $runID
	    }

	    if { $pafFileIPVersion > 0 } {

		set msg "$OBdata($ob_id,$i,tplName) (\#$i): the version of the IP used to create the setup files for $pafFileIPVersion of the pawprints is incorrect (must be $IPVersion.xx, found $full_ip)!"

		# For new runs
		if { $period >= $pVersion } {
		    #error:  the latest IP version MUST be used:
		    if { $obRunMode == "s" } {
			issueError $ob_id $msg $runID
		    } else {
			issueWarning $ob_id $msg
		    }
		} elseif { $period == 98 && [string range $proposal 0 1] == 2 } {
		    issueWarning $ob_id $msg
		} elseif { $period < $pVersion } {
		    issueWarning $ob_id $msg
		}

	    }

	}; # end of loop on templates

    }; # end of loop on OBs

} ;# End proc checkVIRCAMLevel4
