# File: exec_VIRCAM.tcl
# Purpose: computation of execution/exposure time for VIRCAM OBs
# Author: Stephane Marteau

# $Id: exec_VIRCAM.tcl 328278 2019-11-29 14:00:38Z smarteau $

#
# Define global parameters for use within following tests
#

set execCommon(PRESET)           120
set execCommon(PRESET_QUICK)     420
set execCommon(SLEW_OHEAD)        20
set execCommon(SLEW_OHEAD_QUICK) 320
set execCommon(GS_HANDLING)        3
set execCommon(START_AG)           5
set execCommon(START_AO)          45
set execCommon(READOUT)            2
set execCommon(WRITE_TO_DISK)      4
set execCommon(JITTER)             8
set execCommon(PAWPRINT_CHANGE)   15
set execCommon(SKY_OFFSET)        45
set execCommon(FILE_MERGE)         5


set execCommon(customPresetTime) 0
set execCommon(acqName) ""
set execCommon(AG_ENABLED) ""
set execCommon(AG_CONFIRM) ""
set execCommon(AO_ENABLED) ""
set execCommon(AO_CONFIRM) ""

# OB data - global array containing values of all needed parameters for all templates
# of all OBs.
#
# OBdata(<ob_id>,<tpl_idx>,<keyword>) = <value>
#

global OBdata
array set OBdata {}


# computeDistance
#
# Procedure: computes the angular distance between two points
#            of the sky
#
# inputs:
#   - ra1 : right ascension of the first point, given in arcseconds
#   - ra2 : right ascension of the second point, given in arcseconds
#   - dec1: declination of the first point, given in arcseconds
#   - dec2: declination of the second point, given in arcseconds
#
# output:
#  returns the angular distance in arcseconds
#
proc computeDistance { ra1 ra2 dec1 dec2 } {

    # Conversion factor from arcseconds to radians
    set arcsec2rad [expr 3.14159265358979 / (3600. *180.)]

    # convert input coordinates to radians
    set ra1 [expr $ra1 * $arcsec2rad]
    set ra2 [expr $ra2 * $arcsec2rad]
    set dec1 [expr $dec1 * $arcsec2rad]
    set dec2 [expr $dec2 * $arcsec2rad]

    # Compute angular distance:
    # cos(theta) = sin(dec1)*sin(dec2) + cos(dec1)*cos(dec2)*cos(ra2-ra1)
    #

    set t [expr sin($dec1)*sin($dec2) + cos($dec1)*cos($dec2)*cos($ra2-$ra1)]
    set dist [expr acos($t)]

    # Convert to arcseconds
    set dist [expr abs($dist / $arcsec2rad)]

    return $dist
}

# getPresetTime
#
# Procedure: computes the time needed to go from one telescope
# preset to another, taking into account the distance between
# the two pointings.
#
# inputs:
#   - ob1, ob2: OB objects
#
# output:
#  returns the time overhead in seconds
#
proc getPresetTime { ob1 ob2 } {
    global execCommon

    set type1 [evmType $ob1]
    set type2 [evmType $ob2]

    if { $type1 == "C" && $type2 == "O" } {
	return $execCommon(PRESET)
    } elseif { $type1 == "O" && $type2 == "C" } {
	return 0
    }

    # Targets from both OBs
    set target1 [evmTarget $ob1]
    set target2 [evmTarget $ob2]

    # Get right ascension + declination from target packages
    set ra1 [expr [evmRightAscension $target1 marcsec] / 1000.]
    set dec1 [expr [evmDeclination $target1 marcsec] / 1000.]
    set ra2 [expr [evmRightAscension $target2 marcsec] / 1000.]
    set dec2 [expr [evmDeclination $target2 marcsec] / 1000.]

    # Compute angular distance between targets
    set distance [computeDistance $ra1 $ra2 $dec1 $dec2]

    set slewSpeed 3600 ; # slew speed = 1 deg/s = 3600 arcsec/s

    # Overhead =  slew time + execCommon(SLEW_OHEAD) (execCommon(SLEW_OHEAD) is added later in 
    # later in getAcquisitionTime for a "quick" tpl
    return [expr $execCommon(SLEW_OHEAD) + $distance/$slewSpeed]
}


# parseOB
#
# Procedure: parses an OB and stores its contents in the global
# table 'OBdata'.
#
# inputs:
#   - ob: OB object (Java handle)
#
# output:
#   contents of OB are stored in global table 'OBdata'
#
proc parseOB { ob } {

    global OBdata scriptingBox

    set ob_id [evmID $ob]

    # List of all templates (Java handles)
    set templates [evmTemplates $ob]

    # template counter (acquisition=0)
    set index 0

    # Loop on all templates
    foreach temp $templates {

	# Parse all valid templates (verification required to
	# catch exception for 1st template of CalBlocks which is always NULL)
	if { [evmValid $temp] } {

	    # Record template name and type
	    #set OBdata($ob_id,$index,tplType) [evmTemplateType $temp]
	    set OBdata($ob_id,$index,tplName) [evmName $temp]

	    # Parse individual template
	    parseTemplate $temp $ob_id $index

	    incr index ; # template counter: add 1
	}

    }

    # Store how many templates this OB has
    set OBdata($ob_id,nbTpl) $index

    # Extract some keywords from first attached PAF file of every template
    # Used with acq_tile acquisition (see getExpoExectime)
    for { set k 0 } { $k < $index } { incr k } {

	if { $OBdata($ob_id,$k,SEQ.REF.FILE1) == "" } { continue }

	set pafTable [evmParsePAF $OBdata($ob_id,$k,SEQ.REF.FILE1)]
	set OBdata($ob_id,$k,TEL.AO.START) [evmGetPAFParameter $pafTable "TEL.AO.START"]
	set OBdata($ob_id,$k,TEL.AG.START) [evmGetPAFParameter $pafTable "TEL.AG.START"]

    }

    return
}


# Procedure: parseTemplate
# arguments:
#  - template: EVM template object
#
# returns: nothing (global array is used)
#
# Parses a template and puts the parameters into
# a global array
#
proc parseTemplate { template ob_id idx } {

    # Global array in which parameters are stored
    global OBdata

    # List of keywords to retrieve + default values to assign in the code
    # when they do not exist in the given template.
    set keywordList { "DET1.DIT" "-1" \
                      "DET1.NDIT" "-1" 
		      "INS.FILTER.NAME" "not-found" \
                      "INS.FILTER.DARK" "not-found" \
		      "SEQ.TILE.ID" "" \
		      "SEQ.JITTER.ID" "" \
		      "SEQ.JITTER.NJITTER" "" \
                      "SEQ.USTEP.ID" "" \
		      "SEQ.NESTING" "" \
		      "TEL.AG.START" "-1" \
		      "TEL.AO.START" "-1" \
		      "TEL.AG.CONFIRM" "-1" \
		      "TEL.AO.CONFIRM" "-1" \
		      "TEL.AG.GUIDESTAR" "-1" \
		      "TEL.AO.PRIORITY" "-1" \
		      "DET1.DIT" "-1" \
		      "DET1.NDIT" "-1" \
		      "SEQ.NEXPO" "-1" \
		      "SEQ.TILE.ID" "" \
		      "SEQ.JITTER.ID" "" \
		      "SEQ.JITTER.NJITTER" "" \
		      "SEQ.USTEP.ID" "" \
		      "SEQ.NESTING" "" \
		      "TEL.AG.START" "-1" \
		      "TEL.AO.START" "-1" \
		      "TEL.AG.CONFIRM" "-1" \
		      "TEL.AO.CONFIRM" "-1" \
		      "TEL.AG.GUIDESTAR" "-1" \
		      "TEL.AO.PRIORITY" "-1" \
		      "TEL.TARG.ALPHA" "" \
		      "TEL.TARG.DELTA" "" \
		      "SEQ.REF.FILE1" "" \
	    	      "SEQ.SKYJITTER.NJITTER" "" \
	              "SEQ.DET.DITFIXED" ""    }

    # Loop on all keywords to retrieve
    foreach {key def} $keywordList {

	# Retrieve value of keyword from template, assign default value if not found
	set OBdata($ob_id,$idx,$key) [getParamValue $template $key $def]

    }

    return
}


# procedure: getAcquisitionTime
# arguments:
#  - ob_id: ID of the OB containing the template
#  - idx: index of the template within the OB
#
# returns: execution time for acquisition templates of VIRCAM
#
proc getAcquisitionTime { ob_id idx } {

    # Global variables
    global execCommon OBdata

    # Store name of acquisition template into global array
    set templateName $OBdata($ob_id,$idx,tplName)
    set execCommon(acqName) $templateName

    # Preset time: if we compute the slew time then it has to be compared with the
    # time required to switch filters between OBs. Largest overhead wins.
    if { $execCommon(acqName) == "VIRCAM_img_acq_quick" } {
	if { $execCommon(customPresetTime) != 0 } {
	    #remove the slew overhead and add the slew overhead value for a 'quick' tmpl
	    set execCommon(customPresetTime) [expr $execCommon(customPresetTime) - $execCommon(SLEW_OHEAD) + $execCommon(SLEW_OHEAD_QUICK)]
	    set executionTime [max $execCommon(customPresetTime) $execCommon(FILTER_CHANGE_BETWEEN_OBs)]
	} else {   		
	    set executionTime $execCommon(PRESET_QUICK)   
	}
    } else {
	if { $execCommon(customPresetTime) != 0 } {
	    set executionTime [max $execCommon(customPresetTime) $execCommon(FILTER_CHANGE_BETWEEN_OBs)]
	} else {   		
	    set executionTime $execCommon(PRESET)   
	}
    }


    # Figure out extra overheads according to flags
    # in acquisition -- will be used for science templates
    set execCommon(AG_ENABLED) $OBdata($ob_id,$idx,TEL.AG.START)
    set execCommon(AO_ENABLED) $OBdata($ob_id,$idx,TEL.AO.START)

    set execCommon(AG_CONFIRM) $OBdata($ob_id,$idx,TEL.AG.CONFIRM)
    set execCommon(AO_CONFIRM) $OBdata($ob_id,$idx,TEL.AO.CONFIRM)

    set execCommon(AO_PRIORITY) $OBdata($ob_id,$idx,TEL.AO.PRIORITY)


    # Additional overheads when using the autoguider
    if { $OBdata($ob_id,$idx,TEL.AG.START) == "T" } {

	if { $OBdata($ob_id,$idx,TEL.AG.GUIDESTAR) == "CATALOGUE" } {

	    set executionTime [expr $executionTime + $execCommon(GS_HANDLING) \
		+ $execCommon(START_AG) ]

	} else {

	    set executionTime [expr $executionTime + $execCommon(START_AG) ]

	}
    }

    # Additional overheads when using AO
    if { $OBdata($ob_id,$idx,TEL.AO.START) == "T" } {

	if { $OBdata($ob_id,$idx,TEL.AG.GUIDESTAR) == "CATALOGUE" } {

	    set executionTime [expr $executionTime + $execCommon(GS_HANDLING) \
		+ $execCommon(START_AO) ]

	} else {

	    set executionTime [expr $executionTime + $execCommon(START_AO) ]

	}
    }

    # Return the execution time for the acquisition
    return $executionTime

}


# Procedure:getFilterChangeTime
#
# Returns the time needed to change from one filter to another one.
# This time depends on the relative position of the 2 filters in the 
# VIRCAM filter wheel.
#
# Arguments:
#  - filter1: name of current filter
#  - filter2: name of filter to switch to...
#
# Returns:
#  time to switch from filter1 to filter2 (in seconds)
#
proc getFilterChangeTime { filter1 filter2 } {

    # Overhead table (nb of positions to skip --> time in seconds)
    set time(0) 0
    set time(1) 20.5
    set time(2) 27
    set time(3) 33
    set time(4) 40

    # ORDERED List of VIRCAM filters
    set filterList {SUNBLIND NB980 H Ks J Y Z NB118}

    if { [set idx1 [lsearch -exact $filterList $filter1]] == -1 || \
	[set idx2 [lsearch -exact $filterList $filter2]] == -1 } {

	return 0

    }

    set nbFilters [llength $filterList]

    # Since this is a wheel we compute the separation from both ends
    # of the list and take the "shortest path"
    set separation1 [expr abs($idx2 - $idx1)]
    set separation2 [expr $nbFilters - $separation1]
    set separation [min $separation1 $separation2]

    # Get the overhead from the above table
    return $time($separation)

}


# Procedure: getFilterChangeOverhead
# arguments:
#  - ob_id : OB ID
#  - tpl_idx: index of template
#
# returns: overhead due to change of filters (in seconds)
#
# There is an overhead in between two templates whenever the
# last filter of template 1 is not the same as the first
# filter of template 2.
# Takes into account the relative positions of filters to
# compute the overhead.
# If one of the two filters passed as argument is not in the list
# of known filters, the returned value is zero.
#
proc getFilterChangeOverhead { ob_id tpl_idx } {

    global execCommon OBdata

    set overhead 0

    # List of filters used in templates 1 and 2
    set filterTemplate $OBdata($ob_id,$tpl_idx,INS.FILTER.NAME)

    if { $filterTemplate == "not-found" } {
	set filterTemplate $OBdata($ob_id,$tpl_idx,INS.FILTER.DARK)
    }

    # Last filter(s) used (as stored in global array)
    set lastFilter [evmSystemParameterValue "INS.FILTER.NAME"]

    if { $lastFilter == "not-found" } {
	set lastFilter [evmSystemParameterValue "INS.FILTER.DARK"]
    }

    # Check that the 2 templates DO have a list of filters...
    # ... then get the LAST filter of template1 and the FIRST filter
    # of template2
    if {$filterTemplate != "not-found" && \
	    $lastFilter != "not-found" && \
	[set filter1 [lindex $lastFilter end]] != [set filter2 [lindex $filterTemplate 0]] } {

	# Get the associated overhead
	set overhead [getFilterChangeTime $filter1 $filter2 ]

    }

    return $overhead

}

# computeFilterChangeBetweenOBs
#
# Procedure: compute the overhead linked to switching filters in between
# OBs.
# 
# inputs:
#   - id1, id2: IDs of concerned OBs
#
# output:
#   sets a global variable with the overhead between the 2 OBs in seconds
#
proc computeFilterChangeBetweenOBs { id1 id2 } {

    global execCommon OBdata

    # Find last filter in OB1: start from last template and go back until
    # a filter is found.
    set filter1 ""
    for { set i [expr $OBdata($id1,nbTpl) - 1] } { $i >= 0 } { incr i -1} {

	if { $OBdata($id1,$i,INS.FILTER.NAME) != "" } {
	    set filter1 [lindex $OBdata($id1,$i,INS.FILTER.NAME) end]
	    break
	}

    }


    # If no filter found in OB #1, overhead=0 --> exit
    if { $filter1 == "" } { set execCommon(FILTER_CHANGE_BETWEEN_OBs) 0 ; return }

    # Find first filter in OB2: start from first template and go
    # forward until a filter is found.
    set filter2 ""
    for { set i 0 } { $i < $OBdata($id2,nbTpl) } { incr i } {

	if { $OBdata($id2,$i,INS.FILTER.NAME) != "" } {
	    set filter2 [lindex $OBdata($id2,$i,INS.FILTER.NAME) 0]
	    break
	}

    }


    # If no filter found in OB #2, overhead=0 --> exit
    if { $filter2 == "" } { set execCommon(FILTER_CHANGE_BETWEEN_OBs) 0 ; return }

    # Compute the overhead based on the filters that were found
    set execCommon(FILTER_CHANGE_BETWEEN_OBs) [getFilterChangeTime $filter1 $filter2]

    return
}


# Procedure: getExpoExecTime
#
# Top level procedure for exposure/execution time, called from SetExec.
# Computes exposure and execution time for a given science template.
# arguments:
#  - ob_id: ID of the OB that contains the template
#  - tpl_idx: index of the template within the OB
#
# returns: list of exposure + execution time (in seconds)
#
proc getExpoExecTime { ob_id tpl_idx } {

    global execCommon OBdata

    # Reset times for the template
    set exposureTime 0
    set executionTime 0

    # Get template name
    set templateName $OBdata($ob_id,$tpl_idx,tplName)

    # List of science templates.
    set tplList { "VIRCAM_img_obs_tile1" "VIRCAM_img_obs_tile3" "VIRCAM_img_obs_tile6" \
	    	  "VIRCAM_img_obs_tile6sky"  "VIRCAM_img_obs_paw" "VIRCAM_img_cal_std" }

    if { [lsearch -exact $tplList $templateName] != - 1 } {

	# Determine number of tiles
	switch -regexp -- $OBdata($ob_id,$tpl_idx,SEQ.TILE.ID) {

	    {^Tile6}   { set ntiles 6 }
	    {^Tile3}   { set ntiles 3 }
	    {^Tile1|^Single$} { set ntiles 1 }
	    default { set ntiles 1 ; # Case of img_obs_paw, img_cal_std }
	}

	# Determine number of jitter positions
	switch -regexp -- $OBdata($ob_id,$tpl_idx,SEQ.JITTER.ID) {

	    {^Single$}        { set njitter 1 }
	    {^Jitter2[a-z]+$} { set njitter 2 }
	    {^Jitter3[a-z]$}  { set njitter 3 }
	    {^Jitter4[a-z]$}  { set njitter 4 }
	    {^Jitter5[a-z]$}  { set njitter 5 }
	    {^Jitter9[a-z]$}  { set njitter 9 }
	    {^Jitter25[a-z]$} { set njitter 25}
	    {^Jitter30r1$}    { set njitter 30}
	    {^Random$}        { set njitter $OBdata($ob_id,$tpl_idx,SEQ.JITTER.NJITTER) }

	}

	# Determine micro-steps
	switch -- $OBdata($ob_id,$tpl_idx,SEQ.USTEP.ID) {

	    "Single"   {set nusteps 1}
	    "Ustep2"   {set nusteps 2}
	    "Ustep2x2" {set nusteps 4}
	}

	# Number of integrations =
	#  number of tiles * number of jitter positions * number of microsteps
	set nInt [expr $ntiles * $njitter * $nusteps]

	# Number of filters in the template
	set nb_filters [llength $OBdata($ob_id,$tpl_idx,INS.FILTER.NAME)]

	# The DIT list is 'circular' and the list of filters is
	# the driving parameter: if the DIT list is shorter than the list
	# of filters, then the list of DITs is extended to the same length, starting from
	# the beginning of the list again.

	# The loop below extends the DIT list if necessary.
	if { [set nbMissing [expr $nb_filters - [llength $OBdata($ob_id,$tpl_idx,DET1.DIT)]]] > 0 } {

	    set nbDit [llength $OBdata($ob_id,$tpl_idx,DET1.DIT)]

	    set idx 0

	    for {set i 1} {$i <= $nbMissing} {incr i} {

		lappend OBdata($ob_id,$tpl_idx,DET1.DIT) [lindex $OBdata($ob_id,$tpl_idx,DET1.DIT) $idx]

		incr idx

		if {$idx > [expr $nbDit - 1] } { set idx 0 }

	    }

	}

	# The NDIT list is 'circular' and the list of filters is
	# the driving parameter: if the NDIT list is shorter than the list
	# of filters, then the list of NDITs is extended to the same length, starting from
	# the beginning of the list again.

	# The loop below extends the NDIT list if necessary.
	if { [set nbMissing [expr $nb_filters - [llength $OBdata($ob_id,$tpl_idx,DET1.NDIT)]]] > 0 } {

	    set nbNdit [llength $OBdata($ob_id,$tpl_idx,DET1.NDIT)]

	    set idx 0

	    for {set i 1} {$i <= $nbMissing} {incr i} {

		lappend OBdata($ob_id,$tpl_idx,DET1.NDIT) [lindex $OBdata($ob_id,$tpl_idx,DET1.NDIT) $idx]

		incr idx

		if {$idx > [expr $nbNdit - 1] } { set idx 0 }

	    }

	}

	# Sum the exposure time for all individual exposures
	# (loop on the number of filters)
	for {set i 0} {$i < $nb_filters} {incr i} {

	    set dit [lindex $OBdata($ob_id,$tpl_idx,DET1.DIT) $i]
	    set ndit [lindex $OBdata($ob_id,$tpl_idx,DET1.NDIT) $i]

	    set exposureTime [expr $exposureTime + $dit * $ndit]
	}

	#
	# TOTAL EXPOSURE TIME = (number of integrations) x (individual expo. time)
	#
	set exposureTime [expr $nInt * $exposureTime]


	#
	# Computation of execution time (exposure + overheads)
	#

	# How many times are filters changed in the list of filters,
	# and how much overhead this creates
	set filtChangeTime 0

	for {set i 1} {$i < $nb_filters} {incr i} {

	    set filt1 [lindex $OBdata($ob_id,$tpl_idx,INS.FILTER.NAME) [expr $i-1]]
	    set filt2 [lindex $OBdata($ob_id,$tpl_idx,INS.FILTER.NAME) $i]

	    if { $filt2 != $filt1 } {
		set filtChangeTime \
		    [expr $filtChangeTime \
		    + [getFilterChangeTime $filt1 $filt2]]
	    }
	}

	# Compute the time needed to go from the _last_ filter in the list
	# back to the _first_ one, in case the nesting pattern imposes to loop
	# on the list of filters.
	set filtChangeTime2 [getFilterChangeTime \
	    [lindex $OBdata($ob_id,$tpl_idx,INS.FILTER.NAME) end] \
	    [lindex $OBdata($ob_id,$tpl_idx,INS.FILTER.NAME) 0] ]


	# Compute "multiplicative factors" depending on the chosen nesting of loops

	switch -- $OBdata($ob_id,$tpl_idx,SEQ.NESTING) {

	    "FPJME" {
		set filtOverhead $filtChangeTime
		set numPaw [expr $nb_filters * ($ntiles - 1)]

		if { $tpl_idx == 1 } {
		    set nfiltChange [expr $nb_filters - 1]
		} else {
		    set nfiltChange $nb_filters
		}

		set numJit [expr $nb_filters * $ntiles * ($njitter * $nusteps - 1)]
	    }

	    "PFJME" {
		if { $nb_filters > 1 } { 
		    set nfiltChange [expr $ntiles * $nb_filters - 1]
		} else {
		    set nfiltChange 0
		}

		set filtOverhead [expr $ntiles * $filtChangeTime \
		    + ($ntiles - 1) * $filtChangeTime2]
		set numPaw [expr $ntiles - 1]
		set numJit [expr $nb_filters * $ntiles * ($njitter * $nusteps - 1)]
	    }



	    "FJPME" {
		if { $tpl_idx == 1 } {
		    set nfiltChange [expr $nb_filters - 1]
		} else {
		    set nfiltChange $nb_filters
		}

		set filtOverhead $filtChangeTime
		set numPaw [expr $nb_filters * ($ntiles * $njitter - 1)]
		set numJit [expr $nb_filters * $ntiles * $njitter * ($nusteps - 1)]
	    }

	}

	# Add readout overhead for one integration (loop on all filters)
	for {set i 0} {$i < $nb_filters} {incr i} {

	    set dit [lindex $OBdata($ob_id,$tpl_idx,DET1.DIT) $i]
	    set ndit [lindex $OBdata($ob_id,$tpl_idx,DET1.NDIT) $i]

	    set executionTime [expr $executionTime + ($dit + $execCommon(READOUT)) * $ndit]

	}

	# Execution time: (nb integrations) x (exec time for one integration)
	#                + writing to disk
	#                + change filter overhead
	#                + change of pawprint overhead
	#                + jitter overhead
	#
	set executionTime [expr $nInt * $executionTime \
	    + $nInt * $nb_filters * $execCommon(WRITE_TO_DISK) \
	    + $filtOverhead \
	    + $numPaw * $execCommon(PAWPRINT_CHANGE) \
	    + $numJit * $execCommon(JITTER) ]

	# Case of img_obs_tile6sky: compute overhead linked to sky offsets
	if { $templateName == "VIRCAM_img_obs_tile6sky" } {

	    set skyOverhead 0

	    for {set i 0} {$i < $nb_filters} {incr i} {

		set dit [lindex $OBdata($ob_id,$tpl_idx,DET1.DIT) $i]
		set ndit [lindex $OBdata($ob_id,$tpl_idx,DET1.NDIT) $i]

		set skyOverhead [expr $skyOverhead + $dit * $ndit + $execCommon(JITTER)]
	    }		

	    set skyOverhead \
		[expr $njitter * $OBdata($ob_id,$tpl_idx,SEQ.SKYJITTER.NJITTER) * $skyOverhead \
		+ $njitter * 2 * $execCommon(SKY_OFFSET) ]

	    set executionTime [expr $executionTime + $skyOverhead]

	}

	# Add the overheads linked to use of AG and/or AO.
	# Use the settings stored from the acquisition template, or the PAF file
	# depending on the acq template used.
	set AG_AO_overhead 0

	if { $execCommon(acqName) == "VIRCAM_img_acq_tile" && \
		$execCommon(AO_PRIORITY) != "HIGH"} {

	    if { $execCommon(AO_PRIORITY) == "LOW" } {

		set AG_AO_overhead 0

	    } elseif { $execCommon(AO_PRIORITY) == "NORMAL" } {

		set AG_AO_overhead 0

	    }

	} elseif { $execCommon(acqName) == "VIRCAM_img_acq" || \
		($execCommon(acqName) == "VIRCAM_img_acq_tile" && \
		$execCommon(AO_PRIORITY) == "HIGH")}  {

	    if { $execCommon(acqName) == "VIRCAM_img_acq_tile" } {

		set execCommon(AG_ENABLED) $OBdata($ob_id,$tpl_idx,TEL.AG.START)
		set execCommon(AO_ENABLED) $OBdata($ob_id,$tpl_idx,TEL.AO.START)

	    }

	    switch -- $execCommon(AG_ENABLED) {
		"T" { 		    
		    switch -- $execCommon(AO_ENABLED) {
			"T" { set AG_AO_overhead $execCommon(START_AO) }
			"F" { set AG_AO_overhead $execCommon(START_AG) }
		    }
		}

		"F" {
		    switch -- $execCommon(AO_ENABLED) {
			"T" { set AG_AO_overhead $execCommon(START_AO) }
			"F" { }
		    }
		}
	    }

	}

	# Execution time: add the overheads linked to AG/AO
	set executionTime [expr $executionTime + $AG_AO_overhead]


	# Overheads linked to _interactive_ use of AG/AO

	if {$execCommon(AG_ENABLED) == "T" \
		&& $execCommon(AG_CONFIRM) == "T"} {

	    set executionTime [expr $executionTime + $ntiles * 30]

	}

	if {$execCommon(AO_ENABLED) == "T" \
		&& $execCommon(AO_CONFIRM) == "T"} {

	    set executionTime [expr $executionTime + $ntiles * 30]

	}


    }


    ###
    # DAYTIME CALIBRATIONS
    ###

    if { $templateName == "VIRCAM_img_cal_dark" } {

	set exposureTime 0

	set executionTime 0

	foreach dit $OBdata($ob_id,$tpl_idx,DET1.DIT) ndit $OBdata($ob_id,$tpl_idx,DET1.NDIT) {

	    set executionTime [expr $executionTime + \
		$OBdata($ob_id,$tpl_idx,SEQ.NEXPO) * ( ($dit + $execCommon(READOUT)) * $ndit + $execCommon(WRITE_TO_DISK) + $execCommon(FILE_MERGE)) ]

	}

    }

    if { $templateName == "VIRCAM_img_cal_darkcurrent" } {

	set exposureTime 0

	set executionTime 0

	foreach dit $OBdata($ob_id,$tpl_idx,DET1.DIT) {

	    set executionTime [expr $executionTime + \
		$OBdata($ob_id,$tpl_idx,SEQ.NEXPO) * ( ($dit + $execCommon(READOUT)) * $OBdata($ob_id,$tpl_idx,DET1.NDIT) + $execCommon(WRITE_TO_DISK) + $execCommon(FILE_MERGE)) ]

	}

    }

    if { $templateName == "VIRCAM_img_cal_reset" } {
	set exposureTime 0

	set executionTime [expr $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) * \
	    ( $execCommon(READOUT) * $OBdata($ob_id,$tpl_idx,DET1.NDIT) + $execCommon(WRITE_TO_DISK) + $execCommon(FILE_MERGE)) ]

    }

    if { $templateName == "VIRCAM_img_cal_domeflat" } {

	set exposureTime 0

	set executionTime 0

	foreach dit $OBdata($ob_id,$tpl_idx,DET1.DIT) ndit $OBdata($ob_id,$tpl_idx,DET1.NDIT) {

	    set executionTime [expr $executionTime + \
		$OBdata($ob_id,$tpl_idx,SEQ.NEXPO) * ( ($dit + $execCommon(READOUT)) * $ndit + $execCommon(WRITE_TO_DISK) + $execCommon(FILE_MERGE)) ]

	}

	set nb_filters [llength $OBdata($ob_id,$tpl_idx,INS.FILTER.NAME)]
	# How many times are filters changed in the list of filters,
	# and how much overhead this creates
	set filtChangeTime 0

	for {set i 1} {$i < $nb_filters} {incr i} {

	    set filt1 [lindex $OBdata($ob_id,$tpl_idx,INS.FILTER.NAME) [expr $i-1]]
	    set filt2 [lindex $OBdata($ob_id,$tpl_idx,INS.FILTER.NAME) $i]

	    if { $filt2 != $filt1 } {
		set filtChangeTime \
		    [expr $filtChangeTime \
		    + [getFilterChangeTime $filt1 $filt2]]
	    }	    	    

	}

	set executionTime [expr $executionTime + $filtChangeTime ]
    }


    if { $templateName == "VIRCAM_img_cal_noisegain" } {

	set exposureTime 0

	set executionTime [expr 2 * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) * \
	    ( ($OBdata($ob_id,$tpl_idx,DET1.DIT) + $execCommon(READOUT)) * $OBdata($ob_id,$tpl_idx,DET1.NDIT) + $execCommon(WRITE_TO_DISK) + $execCommon(FILE_MERGE) ) ]	

    }

    if { $templateName == "VIRCAM_img_cal_linearity" } {

	set exposureTime 0
	set executionTime 0

	foreach dit $OBdata($ob_id,$tpl_idx,DET1.DIT) {

	    set executionTime [expr $executionTime + \
		2 * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) * ( ($dit + $execCommon(READOUT)) * $OBdata($ob_id,$tpl_idx,DET1.NDIT) + $execCommon(WRITE_TO_DISK) + $execCommon(FILE_MERGE)) ]

	}

	set nb_dit [llength $OBdata($ob_id,$tpl_idx,DET1.DIT)]
	set executionTime [expr $executionTime + \
	    ($nb_dit + 1) * $OBdata($ob_id,$tpl_idx,SEQ.NEXPO) * \
	    ( ($OBdata($ob_id,$tpl_idx,SEQ.DET.DITFIXED) + $execCommon(READOUT)) * $OBdata($ob_id,$tpl_idx,DET1.NDIT) + $execCommon(WRITE_TO_DISK) + $execCommon(FILE_MERGE) )]


    }


    # Return back exposure + execution times as a list
    return [list $exposureTime $executionTime]

}

# setExecTimeForOneOB
#
# Procedure: computes the exposure and execution times
# for a single OB
#
# inputs:
#   - objRef: Java handle to OB object
#
# output:
#   none. Uses the EVM API to internally set the exposure and
#   execution time to the OB inside P2PP
#
proc setExecTimeForOneOB { objRef } {

    global OBdata scriptingBox execCommon

    # Get OB ID
    set ob_id [evmID $objRef]

    # Get list of templates
    set templates [evmTemplates $objRef]

    set previousTemplate "null"

    # Init variables
    set exposureTime 0
    set executionTime 0

    # Template index
    set tplIdx 0

    # Loop on all templates
    foreach temp $templates {

	# NB: test validity of template (avoids working on a NULL object
	# with calblocks that do not have an acq template)
	if { [evmValid $temp] } {

	    #NB 23Sep2009: the function evmType does not seem to work
	    # properly, resorting to template name to determine acq vs. science
	    set templateName $OBdata($ob_id,$tplIdx,tplName)

	    if { [regexp {acq} $templateName] } {

		# Acquisition template: get execution time
		set executionTime [expr $executionTime + [getAcquisitionTime $ob_id $tplIdx]]
		set previousTemplate $temp

	    } else {

		if { $previousTemplate != "null" } {

		    # Filter change time (in case of filter change)
		    set filterChangeTime [getFilterChangeOverhead $ob_id $tplIdx]
		    set executionTime [expr $executionTime + $filterChangeTime]

		}

		# Other templates: get exposure + execution times
		set expo_exec_times [getExpoExecTime $ob_id $tplIdx]
		set exposureTime [expr $exposureTime + [lindex $expo_exec_times 0]]
		set executionTime [expr $executionTime + [lindex $expo_exec_times 1]]

		set previousTemplate $temp

	    }

	    # We store the value of the last filter used (if any) for later retrieval
	    if { $OBdata($ob_id,$tplIdx,INS.FILTER.NAME) != "not-found" } {
		evmSystemSetParameterValue "INS.FILTER.NAME" $OBdata($ob_id,$tplIdx,INS.FILTER.NAME)
	    }

	    if { $OBdata($ob_id,$tplIdx,INS.FILTER.DARK) != "not-found" } {
		evmSystemSetParameterValue "INS.FILTER.DARK" $OBdata($ob_id,$tplIdx,INS.FILTER.DARK)
	    }

	    incr tplIdx
	}

    }; # end of loop on templates

    # Set execution and exposure times in P2PP
    evmSetExecutionTime $objRef [expr round($executionTime)]
    evmSetExposureTime $objRef [expr round($exposureTime)]

    return
}

# Procedure: setExecTime
# arguments: none
# returns  : nothing
#
# Sets the exposure and execution time in P2PP for the selected OB(s)
#
proc setExecTime {} {

    # Global variables
    global scriptingBox
    global execCommon OBdata
    global compoundList compoundListObjects

    # Initialization of EVM process
    set version [evmInit]

    # initialize the system parameters
    #
    # BOB needs to call this procedure separately,
    # therefore skip it, if we are running in BOB
    if { ![evmInBob] } {
	setSystemParameters
    }

    # Analyze selection of OBs + containers
    # and parse all relevant OBs
    retrieveAndParse

    # Loop over all elements in the compound list
    foreach object $compoundListObjects {

	# Get 'object type' (first element of object list)
	set objType [lindex $object 0]

	# If the object type is a list, then we are dealing
	# with a container: the actual type has to be fetched
	# deeper inside the object list
	if { [llength $objType] > 1 } {

	    set objType [lindex [lindex $object 0] 0] ; # Type of container
	    set objRef  [lindex [lindex $object 0] 1] ; # Object reference (Java handle)
	    set objSelected [lindex [lindex $object 0] 2] ; # Is container selected or not

	} else {

	    set objRef [lindex $object 1] ; # Object reference (Java handle)
	    set objSelected [lindex $object 2] ; # Is object part of selection or not

	}	

	if {$objSelected} { set selString "SELECTED" } else { set selString "" }

	switch -regexp -- $objType {

	    {^OB$} {

		#
		# Deal with single (loose) OBs
		#
		set execCommon(customPresetTime) 0
		set execCommon(FILTER_CHANGE_BETWEEN_OBs) 0
		setExecTimeForOneOB $objRef

	    }

	    {GROUP|TIME_LINK} {

		#
		# Groups or Time-Links: deal with OBs individually
		#

		set execCommon(customPresetTime) 0
		set execCommon(FILTER_CHANGE_BETWEEN_OBs) 0

		foreach obj [lindex $object 1] {

		    set type [lindex $obj 0]
		    set ref [lindex $obj 1]
		    set selected [lindex $obj 2]

		    # We skip the OBs that were not explicitly selected
		    if {!$selected} {continue}

		    setExecTimeForOneOB $ref

		}

	    }

	    {CONCATENATION} {

		#
		# Concatenations: OBs executed back-to-back
		#

		set obCounter 0

		foreach obj [lindex $object 1] {

		    incr obCounter

		    set type [lindex $obj 0]
		    set ref [lindex $obj 1]
		    set selected [lindex $obj 2]

		    set execCommon(customPresetTime) 0
		    set execCommon(FILTER_CHANGE_BETWEEN_OBs) 0

		    if { $obCounter > 1 } {

			# Compute preset time from OB <n> --> <n+1>
			set execCommon(customPresetTime) [getPresetTime $previousObRef $ref]

			# Compute filter change overhead between OB <n> and <n+1>
			set previousOBId [evmID $previousObRef]
			set OBId [evmID $ref]
			computeFilterChangeBetweenOBs $previousOBId $OBId

		    }

		    # Compute execution time only for selected OBs
		    if { $selected } {
			setExecTimeForOneOB $ref
		    }

		    set previousObRef $ref
		}

		# Reset special overheads
		set execCommon(customPresetTime) 0
		set execCommon(FILTER_CHANGE_BETWEEN_OBs) 0
		#reset this number as the for loop ends for this concat
		set obCounter 0


	    }

	    default {

	    }
	}

    }

}

# Procedure: setSystemParameters
# arguments: none
# returns  : nothing
#
# Sets the system parameters
#
proc setSystemParameters {} {

    # reset the system parameters
    evmSystemResetParameters { "INS.FILTER.NAME" "INS.FILTER.DARK" }

    # If running in BOB, then let the system parameters reflect
    # the state of the real system.
    if { [evmInBob] } {

	# set INS.FILT.NAME
	set value [bobDbRead "<alias>FILT.status0"]
	if { $value == "" } { set value "not-found" }
	evmSystemSetParameterValue "INS.FILTER.NAME" $value

	# SUNBLIND
	evmSystemSetParameterValue "INS.FILTER.DARK" "SUNBLIND"

    }
}

# Call to the main procedure
setExecTime






